"""
Functions related to the technological stock
"""
from energy_demand.technologies import tech_related

class TechStock(object):
    """Class of a technological stock of a year of the residential model

    The main class of the residential model.
    """
    def __init__(
            self,
            name,
            technologies,
            base_yr,
            curr_yr,
            fueltypes,
            temp_by,
            temp_cy,
            t_base_heating_by,
            potential_enduses,
            t_base_heating_cy,
            enduse_technologies
        ):
        """Constructor of technologies for residential sector

        Arguments
        ----------
        name : str
            Name of technology stock
        technologies : dict
            All technologies and their properties
        base_yr : int
            Base year
        curr_yr : int
            Base year
        fueltypes : dict
            Fueltypes
        temp_by : array
            Base year temperatures
        temp_cy : int
            Current year temperatures
        t_base_heating_by : float
            Base temperature for heating
        potential_enduses : list
            Enduses of technology stock
        t_base_heating_cy : float
            Base temperature current year
        enduse_technologies : list
            Technologies of technology stock

        Notes
        -----
        - The shapes are given for different enduse as technology may be used
          in different enduses and either a technology specific shape is
          assigned or an overall enduse shape
        """
        self.name = name

        self.stock_technologies = create_tech_stock(
            technologies,
            base_yr,
            curr_yr,
            fueltypes,
            temp_by,
            temp_cy,
            t_base_heating_by,
            t_base_heating_cy,
            potential_enduses,
            enduse_technologies)

    def get_tech(self, name, enduse, sector):
        """Get technology of technology stock

        Arguments
        ----------
        name : str
            Name of technology to get
        enduse : str
            Enduse of technology
        sector : str
            Sector of technology
        """
        return self.stock_technologies[(name, enduse, sector)]

    def get_tech_attr(self, enduse, sector, name, attribute_to_get):
        """Get a technology attribute from a technology
        object stored in a list

        Arguments
        ----------
        enduse : string
            Enduse to read technology specified for this enduse
        name : string
            List with stored technologies
        attribute_to_get : string
            Attribute of technology to get

        Return
        -----
        tech_attribute : attribute
            Technology attribute
        """
        tech_object = self.stock_technologies[(name, enduse, sector)]

        attribute_value = getattr(tech_object, attribute_to_get)

        return attribute_value

    def add_tech(
            self,
            name,
            enduse,
            tech_obj
        ):
        """Add technology to technology stock

        Arguments
        ---------
        tech_obj : object
            Technology object
        """
        self.stock_technologies[(name, enduse)] = tech_obj

def create_tech_stock(
        technologies,
        base_yr,
        curr_yr,
        fueltypes,
        temp_by,
        temp_cy,
        t_base_heating_by,
        t_base_heating_cy,
        enduses,
        enduse_technologies
    ):
    """Create technologies and add to dict with key_tuple

    Arguments
    ----------
    technologies : dict
        All technology assumptions
    lookups : dict
        Lookups
    temp_by : array
        Base year temperatures
    temp_cy : int
        Current year temperatures
    t_base_heating_by : float
        Base temperature for heating
    t_base_heating_cy : float
        Base temperature current year
    enduses : list
        Enduses of technology stock
    enduse_technologies : list
        Technologies of technology stock
    """
    stock_technologies = {}

    for enduse in enduses:
        for sector in enduse_technologies[enduse]:
            for technology in enduse_technologies[enduse][sector]:

                if technologies[technology].tech_type == 'placeholder_tech':
                    pass
                else:
                    tech_obj = Technology(
                        name=technology,
                        tech_type=technologies[technology].tech_type,
                        fueltype_str=technologies[technology].fueltype_str,
                        eff_achieved=technologies[technology].eff_achieved,
                        diff_method=technologies[technology].diff_method,
                        eff_by=technologies[technology].eff_by,
                        eff_ey=technologies[technology].eff_ey,
                        year_eff_ey=technologies[technology].year_eff_ey,
                        market_entry=technologies[technology].market_entry,
                        tech_max_share=technologies[technology].tech_max_share,
                        base_yr=base_yr,
                        curr_yr=curr_yr,
                        temp_by=temp_by,
                        temp_cy=temp_cy,
                        t_base_heating_by=t_base_heating_by,
                        t_base_heating_cy=t_base_heating_cy,
                        description=technologies[technology].description)

                    stock_technologies[(technology, enduse, sector)] = tech_obj

            # ----------------------------------------------------------------
            # Add for every enduse placeholder technologies for every fueltype
            # ----------------------------------------------------------------
            for fueltype_str in fueltypes:

                placeholder_technology = 'placeholder_tech__{}'.format(fueltype_str)

                # Placeholder technology
                tech_obj = Technology(
                    name=placeholder_technology,
                    tech_type='placeholder_tech',
                    fueltype_str=fueltype_str)

                stock_technologies[(placeholder_technology, enduse, sector)] = tech_obj

    return stock_technologies

class Technology(object):
    """Technology Class

    Arguments
    ----------
    name : str
        The name of a technology
    tech_type : str
        Technology type
    fueltype_str : str
        Fueltype given as string
    eff_achieved : float
        Percentage of how much the potential efficiency improvement is reaslied
    diff_method : str
        How the diffusion occurs (sigmoid or linear)
    eff_by : float
        Base year efficiency
    eff_ey : float
        Future end year efficiency
    year_eff_ey : float
        Year when the efficiency `eff_ey` is fully realised
    market_entry : float
        Year when technology comes on the market
    tech_max_share : float
        Maximum theoretical penetration of technology
    base_yr : float
        Base year
    curr_yr : float
        Current year
    fueltypes : dict
        Fueltype
    temp_by : array
        Temperatures of base year
    temp_cy : array
        Temperatures of current year
    t_base_heating_by : float
        Base temperature for heating
    t_base_heating_cy : float
        Base temperature current year
    description : str
        Technology description

    Notes
    -----
    * Technologies only coming on the market in the future can be defined by
      defining a future market entry year
    """
    def __init__(
            self,
            name,
            tech_type=None,
            fueltype_str=None,
            eff_achieved=None,
            diff_method=None,
            eff_by=None,
            eff_ey=None,
            year_eff_ey=None,
            market_entry=None,
            tech_max_share=None,
            base_yr=None,
            curr_yr=None,
            temp_by=None,
            temp_cy=None,
            t_base_heating_by=None,
            t_base_heating_cy=None,
            description=''
        ):
        """Contructor
        """
        self.name = name
        self.tech_type = tech_type
        self.description = description
        self.fueltype_str = fueltype_str
        self.fueltype_int = tech_related.get_fueltype_int(fueltype_str)
        self.eff_achieved = eff_achieved
        self.diff_method = diff_method
        self.market_entry = market_entry
        self.tech_max_share = tech_max_share

        if tech_type == 'placeholder_tech':
            self.eff_by = 1.0
            self.eff_cy = 1.0
        elif tech_type == 'heat_pump':
            self.eff_by = tech_related.calc_hp_eff(
                temp_by,
                eff_by,
                t_base_heating_by)

            self.eff_cy = tech_related.calc_hp_eff(
                temp_cy,
                tech_related.calc_eff_cy(
                    base_yr,
                    curr_yr,
                    eff_by,
                    eff_ey,
                    year_eff_ey,
                    self.eff_achieved,
                    self.diff_method),
                t_base_heating_cy)
        else:
            self.eff_by = eff_by
            self.eff_cy = tech_related.calc_eff_cy(
                base_yr,
                curr_yr,
                eff_by,
                eff_ey,
                year_eff_ey,
                self.eff_achieved,
                self.diff_method)

    def set_tech_attr(self, attribute_to_set, value_to_set):
        """Set a technology attribute

        Arguments
        ----------
        attribute_to_set : str
            Attribue to set
        value_to_set : any
            Value to set
        """
        setattr(self, attribute_to_set, value_to_set)



"""Contains very basic functions
"""
import os
import shutil
import numpy as np
from pyproj import Proj, transform

def get_result_paths(folder_path):
    """Joins results subfolders to ``folder_path`` and returns a dict

    Arguments
    ---------
    folder_path : str
        Path of scenario run results

    Returns
    -------
    path_dict : dict
        Dict with all sub_folder paths
    """
    path_dict = {
        'data_results': folder_path,
        'data_results_model_run_pop': os.path.join(folder_path, 'model_run_pop'),
        'data_results_validation': os.path.join(folder_path, 'PDF_validation'),
        'data_results_model_run_results_txt': os.path.join(folder_path, 'model_run_results_txt'),
        'data_results_PDF': os.path.join(folder_path, 'PDF_results')}

    return path_dict

def dict_depth(dictionary):
    """Get depth of nested dict
    """
    if isinstance(dictionary, dict):
        return 1 + (max(map(dict_depth, dictionary.values())) if dictionary else 0)

    return 0

def test_if_sector(dict_to_test, fuel_as_array=False):
    """Test if a dictionary contains also sector information

    Arguments
    ---------
    dict_to_test : dict
        Dict with info to test

    Returns
    -------
    crit_sector_data : bool
        Criteria wheter nested or not

    Example
    -------
    {0: 23, 1: 3434}} --> False
    {'sector': {0: 23, 1: 3434}} --> True
    """
    get_dict_depth = dict_depth(dict_to_test)

    if fuel_as_array: # if given as array, one level less
        if get_dict_depth == 1:
            crit_sector_data = False
        elif get_dict_depth == 2:
            crit_sector_data = True
    else:
        if get_dict_depth == 2:
            crit_sector_data = False
        elif get_dict_depth == 3:
            crit_sector_data = True

    return crit_sector_data

def round_down(num, divisor):
    """Round down
    """
    return num - (num%divisor)

def get_all_folders_files(path):
    """Return all folders and file names in a list

    Input
    -----
    path : str
        Path to folder

    Returns
    --------
    all_folders : list
        All folders in a folder
    filenames : list
        All file names in a list
    """
    folders_walk = os.walk(path)
    for root, dirnames, filenames in folders_walk:
        all_folders = list(dirnames)
        #all_files = list(filenames)
        break

    return all_folders #, all_files

def assign_array_to_dict(array_in, regions):
    """Convert array to dict with same order as region list

    Input
    -----
    regions : list
        List with specific order of regions
    array_in : array
        Data array with data like the order of the region list

    Returns
    -------
    dict_out : dict
        Dictionary of array_in
    """
    dict_out = {}
    for r_idx, region in enumerate(regions):
        dict_out[region] = array_in[r_idx, 0]

    return dict_out

def get_long_lat_decimal_degrees(reg_centroids):
    """Project coordinates from shapefile to get
    decimal degrees (from OSGB_1936_British_National_Grid to
    WGS 84 projection).

    Arguments
    ---------
    reg_centroids : dict
        Centroid information read in from shapefile via smif

    Return
    -------
    reg_coord : dict
        Contains long and latidue for every region in decimal degrees

    Info
    ----
    http://spatialreference.org/ref/epsg/wgs-84/
    """
    reg_coord = {}
    for centroid in reg_centroids:
        in_projection = Proj(init="epsg:27700") # OSGB_1936_British_National_Grid
        put_projection = Proj(init="epsg:4326") # WGS 84 projection

        # Convert to decimal degrees
        long_dd, lat_dd = transform(
            in_projection,
            put_projection,
            centroid['geometry']['coordinates'][0], #longitude
            centroid['geometry']['coordinates'][1]) #latitude

        reg_coord[centroid['properties']['name']] = {}
        reg_coord[centroid['properties']['name']]['latitude'] = lat_dd
        reg_coord[centroid['properties']['name']]['longitude'] = long_dd

    return reg_coord

def rmse(predictions, actual_values):
    """Root-mean-square deviation or
    Root-mean-square-erro (RMSE) calculations

    Arguments
    ----------
    predictions : array
        Model prediction (real value)
    actual_values : array
        Moodelled value

    Returns
    -------
    rmse : array
        root-mean-square deviation

    Info
    -----

        Alternative way

        from sklearn.metrics import mean_squared_error
        from math import sqrt

        rms = sqrt(mean_squared_error(y_actual, y_predicted))
    """
    return np.sqrt(((predictions - actual_values) ** 2).mean())

def array_to_dict(result_array, regions):
    """Convert an array with regions to dict
    with region as key

    Arguments
    ---------
    result_array : array
        Results in region_array
    regions : list
        List with all regions (order is the same)

    Returns
    --------
    result_dict : dict
        reg, value
    """
    result_dict = {}
    for reg_array_nr, region in enumerate(regions):
        result_dict[region] = result_array[reg_array_nr]

    return result_dict

def create_folder(path_folder, name_subfolder=None):
    """Creates folder or subfolder

    Arguments
    ----------
    path : str
        Path to folder
    folder_name : str, default=None
        Name of subfolder to create
    """
    if not name_subfolder:
        if not os.path.exists(path_folder):
            os.makedirs(path_folder)
    else:
        path_result_subolder = os.path.join(path_folder, name_subfolder)
        if not os.path.exists(path_result_subolder):
            os.makedirs(path_result_subolder)

def delete_folder(path_folder):
    """Delete folder or subfolder

    Arguments
    ----------
    path : str
        Path to folder
    folder_name : str, default=None
        Name of subfolder to create
    """
    if os.path.exists(path_folder):
        shutil.rmtree(path_folder)

def del_previous_results(path_folder, path_subfolder_keep):
    """Delete all model results from previous model run. Do not
    delete post installation setup files

    Arguments
    ---------
    path_folder : str
        Path to results of model run
    path_subfolder_keep : str
        Path of subfolder which must not be deleted
    """
    if os.path.exists(path_folder):
        all_files_and_folders = os.listdir(path_folder)

        # Iterate folders in data folders
        for entry in all_files_and_folders:
            path_subfolder = os.path.join(path_folder, entry)

            # Do not deleted post installation files
            if path_subfolder != path_subfolder_keep:
                shutil.rmtree(path_subfolder)
    else:
        pass

def del_previous_setup(path_folder):
    """Delete all model results from previous model run

    Arguments
    ---------
    path_folder : str
        Path to results of model run
    """
    if os.path.exists(path_folder):
        shutil.rmtree(path_folder, ignore_errors=True)
    else:
        pass

def del_file(path_file):
    """Delete all model results from previous model run

    Arguments
    ---------
    path_folder : str
        Path to results of model run
    """
    if os.path.isfile(path_file):
        os.remove(path_file)
    else:
        pass

def get_month_from_string(month_string):
    """Convert string month to int month with Jan == 1

    Arguments
    ----------
    month_string : str
        Month given as a string

    Returns
    --------
    month : int
        Month as an integer (jan = 1, dez = 12)
    """
    if month_string == 'Jan':
        month = 1
    elif month_string == 'Feb':
        month = 2
    elif month_string == 'Mar':
        month = 3
    elif month_string == 'Apr':
        month = 4
    elif month_string == 'May':
        month = 5
    elif month_string == 'Jun':
        month = 6
    elif month_string == 'Jul':
        month = 7
    elif month_string == 'Aug':
        month = 8
    elif month_string == 'Sep':
        month = 9
    elif month_string == 'Oct':
        month = 10
    elif month_string == 'Nov':
        month = 11
    elif month_string == 'Dec':
        month = 12

    return int(month)

def get_month_from_int(month_int):
    """Convert inger month to string month with Jan == 1

    Arguments
    ---------
    month_int : str
        Month given as a integer

    Returns
    --------
    month : int
        Month as an integer (jan = 1, dez = 12)
    """
    if month_int == 1:
        month_str = 'Jan'
    elif month_int == 2:
        month_str = 'Feb'
    elif month_int == 3:
        month_str = 'Mar'
    elif month_int == 4:
        month_str = 'Apr'
    elif month_int == 5:
        month_str = 'May'
    elif month_int == 6:
        month_str = 'Jun'
    elif month_int == 7:
        month_str = 'Jul'
    elif month_int == 8:
        month_str = 'Aug'
    elif month_int == 9:
        month_str = 'Sep'
    elif month_int == 10:
        month_str = 'Oct'
    elif month_int == 11:
        month_str = 'Nov'
    elif month_int == 12:
        month_str = 'Dec'

    return str(month_str)

def remove_element_from_list(input_list, element):
    """Remove element in list

    Arguments
    ---------
    input_list : list
        List with elements
    element : any
        Element to remove

    Returns
    -------
    list_new : list
        List where element is removed
    """
    list_new = []
    for i in input_list:
        if i == element:
            _ = 0
        else:
            list_new.append(i)

    return list_new



"""Map UK MetOffice 2015 data to MARIUS grid
"""
import os

import pandas as pd
import numpy as np
from energy_demand.scripts.weather_scripts import weather_scenario
from energy_demand.geography import weather_station_location

def get_closest_station_met_office(
        latitude_reg,
        longitude_reg,
        weather_stations
    ):
    """
    """
    closest_dist = 99999999999

    for station_id in weather_stations:

        lat_to = weather_stations[station_id]["latitude"]
        long_to = weather_stations[station_id]["longitude"]

        dist_to_station = weather_station_location.calc_distance_two_points(
            lat_from=latitude_reg,
            long_from=longitude_reg,
            lat_to=lat_to,
            long_to=long_to)

        if dist_to_station < closest_dist:
            closest_dist = dist_to_station
            closest_id = station_id

    return closest_id


def map_2015_uk_to_MARIUS_grid(
        path_2015_uk_data,
        path_2015_uk_data_remaped,
        path_example_grid_min_datafile_MARIUS
    ):

    # ------------------------------------------------------------
    # Read in processed UK 2015 data per weather stations and data
    # ------------------------------------------------------------

    # Met Office t_min, t_max
    uk_met_office_stations_t_min_array = np.load(os.path.join(path_2015_uk_data, "2015_t_min.npy"))
    uk_met_office_stations_t_max_array = np.load(os.path.join(path_2015_uk_data, "2015_t_max.npy"))
    
    # Convert data into dict with uk met office station ID
    nr_of_uk_met_office_stations = uk_met_office_stations_t_min_array.shape[0]
    
    stations = pd.read_csv(os.path.join(path_2015_uk_data, "2015_stations.csv"))
    df_stations = stations.set_index(('station_id'))
    stations_2015_met_office = df_stations.to_dict('index')

    uk_met_office_stations_t_min = {}
    uk_met_office_stations_t_max = {}
    for station_array_nr in range(nr_of_uk_met_office_stations):
        station_id = stations.get_value(station_array_nr,'station_id')
        uk_met_office_stations_t_min[station_id] = uk_met_office_stations_t_min_array[station_array_nr]
        uk_met_office_stations_t_max[station_id] = uk_met_office_stations_t_max_array[station_array_nr]

    # ------------------------------------------------------------
    # Read in MARIUS grid weather stations
    # ------------------------------------------------------------
    data_list = weather_scenario.get_temp_data_from_nc(path_example_grid_min_datafile_MARIUS, 'tasmin')
    #list_min = weather_scenario.extend_360_day_to_365(data_list, 'tasmin')
    
    print("... reading in marius weather stations")
    #marius_station_coordinates, _ = weather_scenario.write_weather_data(list_min)
    # Coordinates and station id
    stations_marius = []
    print("NUMBE OF ROWS " + str(len(data_list)))
    cnt = 0
    station_id_cnt = 0
    for i  in data_list.index:
        if cnt == 359: #Originally only 360 days
            # Weather station metadata
            station_lon = data_list.get_value(i,'lon')
            station_lat = data_list.get_value(i,'lat')
            station_id = "station_id_{}".format(station_id_cnt)
            stations_marius.append([station_id, station_lat, station_lon]) #ID, latitude, longitude

            # Reset
            station_id_cnt += 1
            cnt = -1
        cnt += 1
    print("... finished read in marius weather stations")

    nr_of_marius_stations = len(stations_marius)

    # ------------------------------------------------------------
    # Map 2015 UK weather data to MARIUS grid
    # ------------------------------------------------------------
    mapped_marius_station_data_t_min = np.zeros((nr_of_marius_stations, 365))
    mapped_marius_station_data_t_max = np.zeros((nr_of_marius_stations, 365))

    # Iterate marius grid stations
    for station_array_nr, row in enumerate(stations_marius):

        # Marius weather station
        #marius_station_id = row[0]
        marius_station_lat = row[1]
        marius_station_lon = row[2]

        # Get closest Met Office weather station
        closest_uk_2015_met_office_station = get_closest_station_met_office(
            latitude_reg=marius_station_lat,
            longitude_reg=marius_station_lon,
            weather_stations=stations_2015_met_office)
        
        # Add station
        mapped_marius_station_data_t_min[station_array_nr] = uk_met_office_stations_t_min[closest_uk_2015_met_office_station]
        mapped_marius_station_data_t_max[station_array_nr] = uk_met_office_stations_t_max[closest_uk_2015_met_office_station]

    # ----------------------------------------------------------
    # Write out weather stations and data
    # ----------------------------------------------------------
    df = pd.DataFrame(stations_marius, columns=['station_id', 'latitude', 'longitude'])

    df.to_csv(os.path.join(path_2015_uk_data_remaped, "stations_2015.csv"), index=False)
    np.save(os.path.join(path_2015_uk_data_remaped, "t_min.npy"), mapped_marius_station_data_t_min)
    np.save(os.path.join(path_2015_uk_data_remaped, "t_max.npy"), mapped_marius_station_data_t_max)

# =========================
# Write 2015 weather station in MARIUS grid
# =========================
path_2015_uk_data = "X:/nismod/data/energy_demand/H-Met_office_weather_data/_complete_meteo_data_all_yrs_cleaned_min_max"
path_2015_uk_data_remaped = "X:/nismod/data/energy_demand/H-Met_office_weather_data/_complete_meteo_data_all_yrs_cleaned_min_max/2015_remapped"
path_example_grid_min_datafile_MARIUS = "X:/nismod/data/energy_demand/J-MARIUS_data/2020/m00d/daily/WAH_m00d_tasmin_daily_g2_2020.nc"

map_2015_uk_to_MARIUS_grid(
    path_2015_uk_data=path_2015_uk_data,
    path_2015_uk_data_remaped=path_2015_uk_data_remaped,
    path_example_grid_min_datafile_MARIUS=path_example_grid_min_datafile_MARIUS)






"""Append result to shapefile
"""
'''import logging
import shapefile

def write_result_shapefile(lad_geometry_shp, out_shape, field_names, csv_results):
    """
    Join result attributes to LAD geography with
    pyhape library

    Arguments
    ---------
    lad_geometry_shp : str
        Path to LAD shapefile
    out_shape : str
        Path to new shapefile
    field_names : list
        list with new attribute field name
    csv_results : list
        list with result dicts

    Info
    -----
    pip install pyshp
    https://github.com/GeospatialPython/pyshp#reading-shapefiles-from-file-like-objects

    http://www.qgistutorials.com/en/docs/performing_table_joins_pyqgis.html
    """
    # Read in our existing shapefile
    lad_geometry_shp_name = lad_geometry_shp[:-3]
    myshp = open(lad_geometry_shp_name + "shp", "rb")
    mydbf = open(lad_geometry_shp_name + "dbf", "rb")

    record = shapefile.Reader(shp=myshp, dbf=mydbf)

    # Create a new shapefile in memory
    writer = shapefile.Writer()

    # Copy over the existing fields
    writer.fields = list(record.fields)

    # --------------
    # Add new fields
    # --------------
    for field_name in field_names:
        writer.field(field_name, "F", decimal=10) #Float

    # Get position of field 'name' 
    position = 0
    for field_name in record.fields[1:]:
        if field_name[0] == 'name': #corresponds to LAD Geocode
            position_field_name = position
            break
        else:
            position += 1

    # --------------------------
    # Join fields programatically
    # --------------------------
    missing_recors = set()

    # Loop through each record, add a column and get results
    for rec in record.records():

        # Get geocode for row
        geo_code = rec[position_field_name]

        # Iterate result entries in list
        for result_per_field in csv_results:

            # Iterate result entries and add
            try:
                result_csv = result_per_field[geo_code]
            except KeyError:
                # No results
                result_csv = 0
                missing_recors.add(geo_code)

            # Add specific fuel result
            rec.append(result_csv)

        # Add the modified record to the new shapefile
        writer.records.append(rec)

    if missing_recors != []:
        logging.warning(
            "No result value for regions '%s' in joining shapefile",
            missing_recors)
    else:
        pass

    # Copy over the geometry without any changes
    writer._shapes.extend(record.shapes())

    # Save as a new shapefile (or write over the old one)
    writer.save(out_shape)
    logging.info("... finished writing shp")
    return'''



"""Generate charts from multiple scenarios
"""
import os
from energy_demand.read_write import read_data
from energy_demand.basic import date_prop
from energy_demand.plotting import plotting_multiple_scenarios
from energy_demand.basic import basic_functions
from energy_demand.basic import lookup_tables
from energy_demand.read_write import data_loader

from energy_demand.plotting import fig_cross_graphs

def process_scenarios(path_to_scenarios, year_to_model=2015):
    """Iterate folder with scenario results and plot charts

    Arguments
    ----------
    path_to_scenarios : str
        Path to folders with stored results
    year_to_model : int, default=2015
        Year of base year
    """
    # -----------
    # Charts to plot
    # -----------
    heat_pump_range_plot = False        # Plot of changing scenario values stored in scenario name
    plot_multiple_cross_charts = True   # Compare cross charts of different scenario
    comparison_year = 2050
    year_to_plot = 2050

    # Delete folder results if existing
    path_result_folder = os.path.join(
        path_to_scenarios, "__results_multiple_scenarios")

    basic_functions.delete_folder(path_result_folder)

    seasons = date_prop.get_season(
        year_to_model=year_to_model)

    model_yeardays_daytype, _, _ = date_prop.get_yeardays_daytype(
        year_to_model=year_to_model)

    lookups = lookup_tables.basic_lookups()

    # Get all folders with scenario run results (name of folder is scenario)
    scenarios = os.listdir(path_to_scenarios)

    # Simulation information is read in from .ini file for results
    path_fist_scenario = os.path.join(path_to_scenarios, scenarios[0])
    enduses, assumptions, regions = data_loader.load_ini_param(
        path_fist_scenario)

    # -------------------------------
    # Iterate folders and get results
    # -------------------------------
    scenario_data = {}
    for scenario in scenarios:
        scenario_data[scenario] = {}

        all_stations = os.listdir(
            os.path.join(path_to_scenarios, scenario))

        _to_igore = [
            'infoparam.txt',
            'model_run_pop',
            'PDF_validation', 'model_run_sim_param.ini']

        for station in all_stations:
            if station not in _to_igore:

                path_to_result_files = os.path.join(
                    path_to_scenarios,
                    scenario,
                    station,
                    'model_run_results_txt')

                scenario_data[scenario] = read_data.read_in_results(
                    path_result=path_to_result_files,
                    seasons=seasons,
                    model_yeardays_daytype=model_yeardays_daytype)
            else:
                pass

    # Create result folder
    basic_functions.create_folder(path_result_folder)

    # -------------------------------
    # Generate plot with heat pump ranges
    # -------------------------------
    if heat_pump_range_plot:

        plotting_multiple_scenarios.plot_heat_pump_chart(
            lookups,
            regions,
            scenario_data,
            fig_name=os.path.join(path_result_folder, "comparison_hp_service_switch_and_lf.pdf"),
            fueltype_str_input='electricity',
            plotshow=True)

    # -------------------------------
    # Compare cross charts for different scenario
    # Ideally only compare two scenario
    # -------------------------------
    if plot_multiple_cross_charts:
        fig_cross_graphs.plot_cross_graphs_scenarios(
            base_yr=2015,
            comparison_year=comparison_year,
            regions=regions,
            scenario_data=scenario_data,
            fueltype_int=lookups['fueltypes']['electricity'],
            fueltype_str='electricity',
            fig_name=os.path.join(path_result_folder, "cross_chart_electricity.pdf"),
            label_points=False,
            plotshow=False)

        fig_cross_graphs.plot_cross_graphs_scenarios(
            base_yr=2015,
            comparison_year=comparison_year,
            regions=regions,
            scenario_data=scenario_data,
            fueltype_int=lookups['fueltypes']['gas'],
            fueltype_str='gas',
            fig_name=os.path.join(path_result_folder, "cross_chart_gas.pdf"),
            label_points=False,
            plotshow=False)

    # -------------------------------
    # Plot total demand for every year in line plot
    # -------------------------------
    plotting_multiple_scenarios.plot_tot_fueltype_y_over_time(
        scenario_data,
        lookups['fueltypes'],
        fueltypes_to_plot=['electricity', 'gas'],
        fig_name=os.path.join(path_result_folder, "tot_y_multiple_fueltypes.pdf"),
        txt_name=os.path.join(path_result_folder, "tot_y_multiple_fueltypes.txt"),
        plotshow=False)

    plotting_multiple_scenarios.plot_tot_y_over_time(
        scenario_data,
        fig_name=os.path.join(path_result_folder, "tot_y_multiple.pdf"),
        plotshow=False)

    # -------------------------------
    # Plot for all regions demand for every year in line plot
    # -------------------------------
    plotting_multiple_scenarios.plot_reg_y_over_time(
        scenario_data,
        fig_name=os.path.join(path_result_folder, "reg_y_multiple.pdf"),
        plotshow=False)

    # -------------------------------
    # Plot comparison of total demand for a year for all LADs (scatter plot)
    # -------------------------------
    plotting_multiple_scenarios.plot_LAD_comparison_scenarios(
        scenario_data,
        year_to_plot=year_to_plot,
        fig_name=os.path.join(path_result_folder, "LAD_multiple.pdf"),
        plotshow=False)

    # -------------------------------
    # Plot different profiles in radar plot (spider plot)
    # -------------------------------
    plotting_multiple_scenarios.plot_radar_plots_average_peak_day(
        scenario_data,
        fueltype_to_model='electricity',
        fueltypes=lookups['fueltypes'],
        year_to_plot=year_to_plot,
        fig_name=os.path.join(path_result_folder))

    plotting_multiple_scenarios.plot_radar_plots_average_peak_day(
        scenario_data,
        fueltype_to_model='gas',
        fueltypes=lookups['fueltypes'],
        year_to_plot=year_to_plot,
        fig_name=os.path.join(path_result_folder))

    # ----------------------
    # Plot peak hour of all fueltypes for different scenario
    # ----------------------
    plotting_multiple_scenarios.plot_tot_y_peak_hour(
        scenario_data,
        fig_name=os.path.join(path_result_folder, "tot_y_peak_h_electricity.pdf"),
        fueltype_str_input='electricity',
        plotshow=False)

    plotting_multiple_scenarios.plot_tot_y_peak_hour(
        scenario_data,
        fig_name=os.path.join(path_result_folder, "tot_y_peak_h_gas.pdf"),
        fueltype_str_input='gas',
        plotshow=False)
    print("Finished processing multiple scenario")
    return

# Generate plots across all scenarios

#process_scenarios(os.path.abspath("C:/Users/cenv0553/ed/results/scen_A"))
#process_scenarios(os.path.abspath("C:/Users/cenv0553/ed/results/scen_B"))
#process_scenarios(os.path.abspath("C:/Users/cenv0553/ed/results/scen_C"))
#process_scenarios(os.path.abspath("C:/Users/cenv0553/ed/results/scen_D"))

#process_scenarios(os.path.abspath("C:/Users/cenv0553/ed/results/Fig_11"))
process_scenarios(os.path.abspath("C:/Users/cenv0553/ed/results/scen_C_Fig12"))

##process_scenarios(os.path.abspath("C:/Users/cenv0553/ed/results/Fig_08_09"))
#process_scenarios(os.path.abspath("C:/Users/cenv0553/ed/results/Fig_13"))  # Color scheme Fig 13
#process_scenarios(os.path.abspath("C:/Users/cenv0553/ed/results/_multiple_TEST"))


import os
import subprocess
from multiprocessing import Pool, cpu_count
import numpy as np

from energy_demand.read_write import read_weather_data

def my_function(simulation_number):  
    print('simulation_number ' + str(simulation_number))

    run_name = '_low'
    all_weather_stations = False

    same_weather_yr = True
    defined_weather_yr = 2015

    run_smif = False 

    # --------------------------
    # Get all weather yrs with data (maybe read existing years from data folder)
    # and select weather yr
    # --------------------------
    path_to_weather_data = "/soge-home/staff/cenv0553/data/_raw_data/A-temperature_data/cleaned_weather_stations_data"
    weather_yrs_stations = read_weather_data.get_all_station_per_weather_yr(path_to_weather_data, min_nr_of_stations=30)
    weather_yrs = list(weather_yrs_stations.keys())
    print("all weather yrs:         " + str(weather_yrs))
    print("Total nr of stations:    " + str(len(weather_yrs)))

    if same_weather_yr:
        weather_yr = defined_weather_yr
    else:
        weather_yr = weather_yrs[simulation_number]

    # --------------------------
    # Select weather station
    # --------------------------
    if all_weather_stations:
        weather_station_cnt = [] #All stationssimulation_number
    else:
        weather_station_cnt = simulation_number

    # Make run name_specifiv
    run_name = "{}_{}".format(run_name, simulation_number)

    # Run energy demand main.py
    if run_smif:
        # Run smif
        #bash_command = "smif -v run ed_constrained_pop-baseline16_econ-c16_fuel-c16"
        #os.system(bash_command)
        pass
    else:
        bash_command = "python energy_demand/energy_demand/main.py {} {} {}".format(run_name, weather_yr, weather_station_cnt)
        os.system(bash_command)
    return 

# WEather years
simulation_number = range(10)

if __name__ == "__main__":
    with Pool(int(cpu_count()/2)) as pool:
        pool.map(
            my_function,
            simulation_number,
            chunksize=1)

'''
for i in range(2):

    print("run " + str(i))

    # Activate virtual environement
    bashCommand = "activate ed"
    os.system(bashCommand)

    # Run smif
    bashCommand = "smif -v run ed_constrained_pop-baseline16_econ-c16_fuel-c16"
    os.system(bashCommand)

    #process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    #output, error = process.communicate()
'''
'''
#import this
from multiprocessing import Pool, cpu_count

#import any other packages 
import numpy as np


def         my_function(simulation_number):  
print('simulation_number')
                return 


simulation_list = [1,2,3,4,5,6,7,8,9,10]

if __name__ == "__main__":
    with Pool(int(cpu_count()/2)) as pool:
         pool.map(my_function,simulation_list,chunksize=1)
#import sh
#sh.cd('C:/Users/cenv0553/ed')

#print(sh.pwd())

#stream = os.popen("cd C:/Users/cenv0553/ed")
'''


"""
"""
import os
import math
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats
import pandas as pd

from energy_demand.plotting import plotting_styles
from energy_demand.basic import basic_functions
from energy_demand.plotting import basic_plot_functions
from energy_demand.plotting import fig_lf

def run(
        path_fig_folder,
        path_plot_fig,
        calc_av_lp_modelled,
        calc_av_lp_real,
        calc_lp_modelled=None,
        calc_lp_real=None,
        plot_peak=False,
        plot_radar=False,
        plot_all_entries=False,
        plot_max_min_polygon=True,
        plotshow=False,
        max_y_to_plot=60,
        fueltype_str=False,
        year=False
    ):
    """Plotting average saisonal loads for each daytype. As an input
    GWh is provided, which for each h is cancelled out to GW.

    https://stackoverflow.com/questions/4325733/save-a-subplot-in-matplotlib
    http://matthiaseisen.com/matplotlib/shapes/reg-polygon/
    """
    fig = plt.figure(
        figsize=basic_plot_functions.cm2inch(14, 25))

    ax = fig.add_subplot(nrows=4, ncols=2)

    plot_nr = 0
    row = -1
    for season in calc_av_lp_modelled:
        row += 1
        col = -1
        for daytype in calc_av_lp_modelled[season]:
            col += 1
            plot_nr += 1

            axes = plt.subplot(4, 2, plot_nr)

            # ------------------
            # Plot average
            # ------------------
            x_values = range(24)

            plt.plot(
                x_values,
                list(calc_av_lp_real[season][daytype]),
                color='black',
                label='av_real or av by',
                linestyle='--',
                linewidth=0.5)

            plt.plot(
                x_values,
                list(calc_av_lp_modelled[season][daytype]),
                color='blue',
                label='av_modelled or av cy',
                linestyle='--',
                linewidth=0.5)

            # --------------
            # Radar plots
            # --------------
            if plot_radar:
                name_spider_plot = os.path.join(
                    path_fig_folder,
                    "spider_{}_{}_{}_{}_.pdf".format(
                        year,
                        fueltype_str,
                        season,
                        daytype))

                plot_radar_plot(
                    list(calc_av_lp_modelled[season][daytype]),
                    name_spider_plot,
                    plot_steps=20,
                    plotshow=False)

            # ------------------
            # Plot every single line
            # ------------------
            if plot_all_entries:
                for entry in range(len(calc_lp_real[season][daytype])):
                    plt.plot(
                        x_values,
                        list(calc_lp_real[season][daytype][entry]),
                        color='grey',
                        markersize=0.5,
                        alpha=0.2)

                    plt.plot(
                        x_values,
                        list(calc_lp_modelled[season][daytype][entry]),
                        color='blue',
                        markersize=0.5,
                        alpha=0.2)

            # ----------
            # Plot max_min range polygons
            # ----------
            if plot_max_min_polygon:

                # ----Draw real
                min_max_polygon = []
                upper_boundary = []
                lower_bdoundary = []

                # Get min and max of all entries for hour
                for hour in range(24):
                    min_y = np.min(calc_lp_real[season][daytype][:, hour], axis=0)
                    max_y = np.max(calc_lp_real[season][daytype][:, hour], axis=0)
                    upper_boundary.append((hour, min_y))
                    lower_bdoundary.append((hour, max_y))

                # create correct sorting to draw filled polygon
                min_max_polygon = fig_lf.order_polygon(upper_boundary, lower_bdoundary)
                #min_max_polygon = create_min_max_polygon_from_lines(reg_load_factor_y)

                polygon = plt.Polygon(
                    min_max_polygon,
                    color='grey',
                    alpha=0.2,
                    edgecolor=None,
                    linewidth=0,
                    fill='True')

                axes.add_patch(polygon)

                # -----Draw modelled
                min_max_polygon = []
                upper_boundary = []
                lower_bdoundary = []

                # Get min and max of all entries for hour
                for hour in range(24):
                    min_y = np.min(calc_lp_modelled[season][daytype][:, hour], axis=0)
                    max_y = np.max(calc_lp_modelled[season][daytype][:, hour], axis=0)
                    upper_boundary.append((hour, min_y))
                    lower_bdoundary.append((hour, max_y))

                # create correct sorting to draw filled polygon
                min_max_polygon = fig_lf.order_polygon(upper_boundary, lower_bdoundary)

                polygon = plt.Polygon(
                    min_max_polygon,
                    color='blue',
                    alpha=0.2,
                    edgecolor=None,
                    linewidth=0,
                    fill='True')

                axes.add_patch(polygon)

            # --------------------
            # Get load shape within season with highest houly load
            # --------------------
            if plot_peak:

                # Get row with maximum hourly value
                day_with_max_h = np.argmax(np.max(calc_lp_real[season][daytype], axis=1))

                plt.plot(
                    x_values,
                    list(calc_lp_real[season][daytype][day_with_max_h]),
                    color='grey',
                    markersize=1.0,
                    label='real_peak or by peak',
                    linestyle='-.',
                    linewidth=0.5)

                # Get row with maximum hourly value
                day_with_max_h = np.argmax(np.max(calc_lp_modelled[season][daytype], axis=1))

                plt.plot(
                    x_values,
                    list(calc_lp_modelled[season][daytype][day_with_max_h]),
                    color='blue',
                    markersize=1.0,
                    label='modelled_peak or cy peak',
                    linestyle='-.',
                    linewidth=0.5)

            # -----------------
            # Axis
            # -----------------
            plt.ylim(0, max_y_to_plot)
            plt.xlim(0, 23)

            # Tight layout
            plt.tight_layout()
            plt.margins(x=0)

            # Calculate RMSE
            rmse = basic_functions.rmse(
                calc_av_lp_modelled[season][daytype],
                calc_av_lp_real[season][daytype])

            # Calculate R_squared
            slope, intercept, r_value, p_value, std_err = stats.linregress(
                calc_av_lp_modelled[season][daytype],
                calc_av_lp_real[season][daytype])

            # Calculate standard deviation
            std_dev_p = np.std(calc_av_lp_real[season][daytype] - calc_av_lp_modelled[season][daytype])
            std_dev_abs = np.std(abs(calc_av_lp_real[season][daytype] - calc_av_lp_modelled[season][daytype]))

            # -----------
            # Labelling
            # -----------
            font_additional_info = plotting_styles.font_info()

            title_info = ('{}, {}'.format(season, daytype))
            plt.text(1, 0.55, "RMSE: {}, R_squared: {}, std: {} (+- {})".format(
                round(rmse, 2),
                round(r_value, 2),
                round(std_dev_p, 2),
                round(std_dev_abs, 2),
                fontdict=font_additional_info))
            plt.title(title_info, loc='left', fontdict=font_additional_info)

    # ------------
    # Plot legend
    # ------------
    plt.legend(
        ncol=1,
        loc=2,
        prop={
            'family': 'arial',
            'size': 5},
        frameon=False)

    # Tight layout
    plt.tight_layout()
    plt.margins(x=0)

    fig.savefig(path_plot_fig)

    if plotshow:
        plt.show()
        plt.close()
    else:
        plt.close()

def plot_radar_plot(
        dh_profile,
        fig_name,
        plot_steps=30,
        plotshow=False
    ):
    """Plot daily load profile on a radar plot

    Arguments
    ---------
    dh_profile : list
        Dh values to plot
    fig_name : str
        Path to save figure

    SOURCE: https://python-graph-gallery.com/390-basic-radar-chart/
    """

    # Get maximum demand
    max_entry = np.array(dh_profile).max()
    max_demand = round(max_entry, -1) + 10 # Round to nearest 10 plus add 10
    max_demand = 120

    nr_of_plot_steps = int(max_demand / plot_steps) + 1

    axis_plots_inner = []
    axis_plots_innter_position = []

    # Inner ciruclar axis
    for i in range(nr_of_plot_steps):
        axis_plots_inner.append(plot_steps*i)
        axis_plots_innter_position.append(str(plot_steps*i))

    data = {'dh_profile': ['testname']}

    for hour in range(24):

        # Key: Label outer circle
        data[hour] = dh_profile[hour]

    # Set data
    df = pd.DataFrame(data)

    # number of variable
    categories=list(df)[1:]

    N = len(categories)

    # We are going to plot the first line of the data frame.
    # But we need to repeat the first value to close the circular graph:
    values=df.loc[0].drop('dh_profile').values.flatten().tolist()

    # Because 0 entry in list is 01:00 time, a value needs to be
    # added for the midnight hour. We therefore add the last entry
    # also to the first position (24:00 value).
    values.insert(0, values[23])

    # What will be the angle of each axis in the plot? (we divide the plot / number of variable)
    angles = [n / float(N) * 2 * math.pi for n in range(N)]
    angles += angles[:1]

    # Initialise the spider plot
    ax = plt.subplot(111, polar=True)

    # Change circula axis
    ax.yaxis.grid(color='lightgrey', linestyle='--', linewidth=0.8, alpha=0.8) # Circular axis
    ax.xaxis.grid(color='lightgrey', linestyle='--', linewidth=0.8, alpha=0.8) # Regular axis

    # Change to clockwise cirection
    ax.set_theta_direction(-1)
    #ax.set_theta_offset(pi/2.0)

    # Set first hour on top
    ax.set_theta_zero_location("N")

    # Draw one axe per variable + add labels labels yet
    plt.xticks(
        angles[:-1],
        categories,
        color='grey',
        size=8)

    # Draw ylabels
    ax.set_rlabel_position(0)
    plt.yticks(
        axis_plots_inner,
        axis_plots_innter_position,
        color="grey",
        size=7)

    # Set limit to size
    plt.ylim(0, max_demand)

    # Smooth lines
    angles_smoothed, values_smoothed = basic_plot_functions.smooth_data(
        angles, values, spider=True)

    # Plot data
    ax.plot(
        angles_smoothed,
        values_smoothed,
        linestyle='--',
        linewidth=0.5)

    # plot data points
    '''ax.scatter(angles, values, color="blue", s=10)'''

    '''line_zeros = np.zeros(len(angles_smoothed))
    ax.fill_between(
        angles_smoothed, #zero coordinates
        line_zeros, # line 1
        values_smoothed, # line 2
        color='blue',
        alpha=0.1)'''

    # Fill below
    ax.fill(
        angles_smoothed, #angles,
        values_smoothed, #values,
        'blue', #b
        alpha=0.1)

    # Save fig
    plt.savefig(fig_name)

    if plotshow:
        plt.show()
        plt.close()
    else:
        plt.close()



"""
"""
from energy_demand.technologies import diffusion_technologies

def test_linear_diff():
    """Testing function 
    """
    expected1 = 1
    expected2 = 1.5
    expected3 = 10
    expected4 = 0

    # call function
    out_value_1 = diffusion_technologies.linear_diff(
        base_yr=2015,
        curr_yr=2020,
        value_start=0.0,
        value_end=1.0,
        yr_until_changed=2020)
    assert out_value_1 == expected1
    out_value_2 = diffusion_technologies.linear_diff(
        base_yr=2015,
        curr_yr=2016.5,
        value_start=1.0,
        value_end=2.0,
        yr_until_changed=2018)
    assert out_value_2 == expected2
    out_value_3 = diffusion_technologies.linear_diff(
        base_yr=2000,
        curr_yr=2100,
        value_start=0,
        value_end=100.0,
        yr_until_changed=3000)
    assert out_value_3 == expected3
    out_value_4 = diffusion_technologies.linear_diff(
        base_yr=2015,
        curr_yr=2015,
        value_start=0,
        value_end=1,
        yr_until_changed=2015)
    assert out_value_4 == expected4

def test_sigmoid_diffusion():
    """testing
    """
    base_yr = 2015
    curr_yr = 2015
    end_yr = 2020
    sig_midpoint = 0
    sig_steepness = 1

    result = diffusion_technologies.sigmoid_diffusion(
        base_yr,
        curr_yr,
        end_yr,
        sig_midpoint,
        sig_steepness)

    assert result == 0


    base_yr = 2015
    curr_yr = 2020
    end_yr = 2020
    sig_midpoint = 0
    sig_steepness = 1

    result = diffusion_technologies.sigmoid_diffusion(
        base_yr,
        curr_yr,
        end_yr,
        sig_midpoint,
        sig_steepness)

    assert result == 1
    
    # ---

    base_yr = 2015
    curr_yr = 2015
    end_yr = 2020
    sig_midpoint = 0
    sig_steepness = 1

    result = diffusion_technologies.sigmoid_diffusion(
        base_yr,
        curr_yr,
        end_yr,
        sig_midpoint,
        sig_steepness)

    assert result == 0    

    # ---

    base_yr = 2015
    curr_yr = 2020
    end_yr = 2025
    sig_midpoint = 0
    sig_steepness = 1

    result = diffusion_technologies.sigmoid_diffusion(
        base_yr,
        curr_yr,
        end_yr,
        sig_midpoint,
        sig_steepness)

    assert result == 0.5



"""
"""
import os
from collections import defaultdict
import numpy as np

def read_weather_station_temp(path_files, years_to_load=[]):
    """Get temperature values. Based on the defined
    years of which the temperature should be loaded,
    it is checked which weather stations occur
    in every year. Only these values are loaded

    Arguments
    ---------
    years_to_load : list
        Years to load temperature data

    Returns
    -------
    """
    container = {}

    all_files = os.listdir(path_files)

    # ----------------------------------------------------------
    #
    # ----------------------------------------------------------
    for file_name in all_files:
        year = int(file_name.split("__")[0])                # year
        station_name = str(file_name.split("__")[1][:-4])   # station_name
        try:
            container[station_name].append(year)
        except KeyError:
            container[station_name] = [year]

    # ---------------------------------------------------------
    # Test for which station names there is data for every year
    # ---------------------------------------------------------
    station_with_data_every_year = []
    for station_name, years in container.items():
        all_years_available = set(years_to_load).issubset(years)

        if all_years_available:
            station_with_data_every_year.append(station_name)

    print("Number of stations with every year")
    print(len(station_with_data_every_year))
    print("Stations with every year")
    print(station_with_data_every_year)
    #raise Exception
    # -------
    stations_in_all_y = defaultdict(dict)
    all_stations = defaultdict(dict)
    for station_name_to_load in station_with_data_every_year:

        for year_to_load in years_to_load:

            for file_name in all_files:
                year = int(file_name.split("__")[0])                # year
                station_name = str(file_name.split("__")[1][:-4])   # station_name

                if year_to_load == year and station_name_to_load == station_name:
                    file_path = os.path.join(path_files, file_name)
                    stations_in_all_y[station_name][year_to_load] = np.loadtxt(file_path, delimiter=',')
                else:
                    file_path = os.path.join(path_files, file_name)
                    all_stations[station_name][year_to_load] = np.loadtxt(file_path, delimiter=',')

    # print info
    print("Info----------------")
    for station_name, years in stations_in_all_y.items():
        print("Number of years in station {}: {}".format(len(list(years.keys())), station_name))
    
    print(" ")
    print("=======================")
    print(" ")
    for station_name, years in all_stations.items():
        print("Number of years in station {}: {}".format(len(list(years.keys())), station_name))
    return dict(stations_in_all_y)

read_weather_station_temp(
    path_files = "//linux-filestore.ouce.ox.ac.uk/mistral/nismod/data/energy_demand/H-Met_office_weather_data/_year_selection_cleaned_2015_14_13",
    years_to_load=[2013,2014,2015]
)


"""The function `EnergyDemandModel` executes all the submodels of the energy demand model

Notes
-----
some observations from @willu47

- suggest moving bulk of EnergyDemandModel class contents out of __init__ method
  and into class methods, and then calling each method in turn
- suggest explictly creating class properties to access results, rather than
  the ``setattr`` magic
- potential for parallelisation of the for-loop around regions for generation
  by dwelling stocks
- suggest splitting ``aggregate_across_all_regions`` into separate methods as each
  of these requires creating a very large array which is then held in memory,
  along with all of the disaggregated results. Ideally, only those aggregatation
  which are required could be run, then written to disk and then removed from
  memory before moving onto the next
- Lots of potential for a map-reduce approach to this, which could be streamlined
  by using vectorised numpy methods consistently throughout the codebase rather
  than for-loops over indexes of the arrays - this will result in memory and
  performance improvements
- Otherwise, the many of the methods are nicely packaged and will be easily
  testable and refactored going forward, so some profiling would help direct
  attention to where further work is most appropriate
"""
import logging
from collections import defaultdict
import numpy as np

try:
    from tqdm import tqdm
except ImportError:
    def tqdm(iterator, *_, **__):
        """Alternative to tqdm, with no progress bar - ignore any arguments after the first
        """
        return iterator

from energy_demand.basic import lookup_tables
import energy_demand.enduse_func as endusefunctions
from energy_demand.geography.region import Region
from energy_demand.geography.weather_region import WeatherRegion
from energy_demand.dwelling_stock import dw_stock
from energy_demand.profiles import load_factors
from energy_demand.profiles import generic_shapes
from energy_demand.basic import demand_supply_interaction
from energy_demand.geography import weather_station_location

class EnergyDemandModel(object):
    """ Main function of energy demand model. All submodels
    are executed here and all aggregation functions of the results

    Arguments
    ----------
    regions : list
        Region names
    data : dict
        Main data container
    assumptions : obj
        Assumptions and calculations based on assumptions
    weather_yr : int
        Year of temperature
    """
    def __init__(
            self,
            regions,
            data,
            criterias,
            assumptions,
            weather_stations,
            weather_yr,
            weather_by
        ):
        """Constructor
        """
        self.curr_yr = assumptions.curr_yr

        # ------------------------
        # Create Dwelling Stock
        # ------------------------
        logging.debug("... generating dwelling stocks")
        lookups = lookup_tables.basic_lookups()

        if criterias['virtual_building_stock_criteria']:
            rs_dw_stock, ss_dw_stock = create_virtual_dwelling_stocks(
                regions, assumptions.curr_yr, data, criterias)

            data['dw_stocks'] = {
                lookups['submodels_names'][0]: rs_dw_stock,
                lookups['submodels_names'][1]: ss_dw_stock,
                lookups['submodels_names'][2]: None}
        else:
            # Create dwelling stock from imported data from newcastle
            data = create_dwelling_stock(
                regions, assumptions.curr_yr, data)

        # Initialise result container to aggregate results
        aggr_results = initialise_result_container(
            assumptions.fueltypes_nr,
            assumptions.reg_nrs,
            assumptions.lookup_enduses,
            assumptions.nr_of_submodels)

        # -------------------------------------------
        # Simulate regions
        # -------------------------------------------
        #_all_closest_weather_stations = []
        for reg_array_nr, region in enumerate(tqdm(regions)):

            print("... Simulate: region %s, simulation year: %s, percent: (%s)",
                region, assumptions.curr_yr, round((100/assumptions.reg_nrs)*reg_array_nr, 2), flush=True)
            #logging.info("... Simulate: region %s, simulation year: %s, percent: (%s)",
            #    region, assumptions.curr_yr, round((100/assumptions.reg_nrs)*reg_array_nr, 2))
            all_submodels = simulate_region(
                region,
                data,
                criterias,
                assumptions,
                weather_stations,
                weather_yr,
                weather_by)
            #_all_closest_weather_stations.append(all_submodels)#weather_region_id
    
            # ---------------------------------------------
            # Aggregate results specifically over regions
            # ---------------------------------------------
            aggr_results = aggregate_single_region(
                assumptions.reg_nrs,
                assumptions.lookup_enduses,
                aggr_results,
                reg_array_nr,
                all_submodels,
                criterias['mode_constrained'],
                assumptions.fueltypes_nr,
                assumptions.enduse_space_heating,
                assumptions.technologies)

        # ---------------------------------------------------
        # Aggregate results for all regions
        # ---------------------------------------------------
        aggr_results = aggregate_across_all_regs(
            aggr_results,
            assumptions.fueltypes_nr,
            data['assumptions'].reg_nrs,
            data['enduses'],
            data['assumptions'],
            criterias,
            assumptions.technologies)

        # Set all keys of aggr_results as self.attributes
        for key_attribute_name, value in aggr_results.items():
            setattr(self, key_attribute_name, value)

def aggregate_across_all_regs(
        aggr_results,
        fueltypes_nr,
        reg_nrs,
        enduses,
        assumptions,
        criterias,
        technologies
    ):
    """Aggregate the regional results into different
    forms used for outputs

    Arguments
    ---------
    aggr_results : dict
        Aggregation container
    fueltypes_nr : int
        Number of fueltypes
    reg_nrs : int
        Number of regions
    enduses : dict
        Submodels and their sectors
    assumptions : dict
        Assumptions
    criterias : dict
        Criteria
    technologies : dict
        Technologies per enduse
    """
    fueltypes = lookup_tables.basic_lookups()['fueltypes']

    # ----------------------------------------------------
    # Aggregate: [fueltype, region, fuel_yh_8760]
    # ----------------------------------------------------
    array_init = np.zeros((fueltypes_nr, reg_nrs, 8760))
    aggr_results['ed_fueltype_regs_yh'] = aggregate_from_full_results(
        assumptions.lookup_enduses,
        array_init,
        aggr_results['ed_enduse_fueltype_regs_yh'],
        time_resolution='8760_h',
        per_region=True,
        per_sector=False,
        per_enduse=False)

    # ----------------------------------------------------
    # Aggregate: [fueltype, fuel_yh_8760]
    # ----------------------------------------------------
    array_init = np.zeros((fueltypes_nr, 365, 24))
    aggr_results['ed_fueltype_national_yh'] = aggregate_from_full_results(
        assumptions.lookup_enduses,
        array_init,
        aggr_results['ed_enduse_fueltype_regs_yh'],
        time_resolution='365_24',
        per_region=False,
        per_sector=False,
        per_enduse=False)

    # ----------------------------------------------------
    # Aggregate: [fueltype, enduse, fuel_yh, 8760]
    # ----------------------------------------------------
    array_init = {}
    for submodel in enduses:
        for enduse in enduses[submodel]:
            array_init[enduse] = np.zeros((fueltypes_nr, 8760))
    aggr_results['tot_fuel_y_enduse_specific_yh'] = aggregate_from_full_results(
        assumptions.lookup_enduses,
        array_init,
        aggr_results['ed_enduse_fueltype_regs_yh'],
        time_resolution='8760_h',
        per_region=False,
        per_sector=False,
        per_enduse=True)

    # ----------------------------------------------------
    # Calculate averaged hour profile per season
    # ----------------------------------------------------
    aggr_results['averaged_h'] = averaged_season_hourly(
        aggr_results['ed_fueltype_regs_yh'],
        fueltypes.values(),
        fueltypes_nr,
        assumptions.seasons,
        reg_nrs)

    # ----------------------------------------------------
    # Regional load factor calculations
    # ----------------------------------------------------
    aggr_results['reg_load_factor_y'], aggr_results['reg_load_factor_yd'] = aggregate_load_factors(
        aggr_results['ed_fueltype_regs_yh'],
        fueltypes_nr,
        reg_nrs)

    # ----------------------------------------------------
    # Unconstrained results
    # np.array of all fueltypes (submodel, region, fueltype, hours)
    # ----------------------------------------------------
    '''aggr_results['results_unconstrained'] = aggregate_result_unconstrained(
        assumptions.nr_of_submodels,
        assumptions.lookup_sector_enduses,
        aggr_results['ed_enduse_fueltype_regs_yh'],
        fueltypes_nr,
        reg_nrs)'''

    # ----------------------------------------------------
    # Generate dict for supply model
    # ----------------------------------------------------
    if criterias['mode_constrained']:
        aggr_results['supply_results'] = demand_supply_interaction.constrained_results(
            aggr_results['results_constrained'],
            #aggr_results['results_unconstrained'],
            aggr_results['results_unconstrained_no_heating'],
            assumptions.submodels_names,
            technologies,
            reg_nrs,
            assumptions.fueltypes_nr)
    else:
        aggr_results['supply_results'] = demand_supply_interaction.unconstrained_results(
            aggr_results['results_unconstrained'],
            assumptions.submodels_names)

    return aggr_results

def aggregate_load_factors(
        ed_fueltype_regs_yh,
        fueltypes_nr,
        reg_nrs
    ):
    """Calculate load factors and aggregate
    """
    reg_load_factor_y = np.zeros((fueltypes_nr, reg_nrs), dtype="float")
    reg_load_factor_yd = np.zeros((fueltypes_nr, reg_nrs, 365), dtype="float")

    # Calculate annual load factors across all enduses
    for fueltype_nr in range(fueltypes_nr):
        for reg_array_nr in range(reg_nrs):

            reg_load_factor_y[fueltype_nr][reg_array_nr] = load_factors.calc_lf_y_8760(
                ed_fueltype_regs_yh[fueltype_nr][reg_array_nr])

            # Calculate average load for every day
            reg_load_factor_yd[fueltype_nr][reg_array_nr] = load_factors.calc_lf_d_8760(
                ed_fueltype_regs_yh[fueltype_nr][reg_array_nr])

    return reg_load_factor_y, reg_load_factor_yd

def get_disaggregated_fuel_of_reg(submodel_names, fuel_disagg, region):
    """Get disaggregated region for every submodel
    for a specific region

    Arguments
    -------
    submodel_names : dict
        Name of all submodels
    fuel_disagg : dict
        Fuel per submodel for all regions
    region : str
        Region

    Returns
    -------
    region_fuel_disagg : dict
        Disaggregated fuel for a specific region
    """
    region_fuel_disagg = {}

    for submodel_name in submodel_names:
        region_fuel_disagg[submodel_name] = fuel_disagg[submodel_name][region]

    return region_fuel_disagg

def simulate_region(
        region,
        data,
        criterias,
        assumptions,
        weather_stations,
        weather_yr,
        weather_by
    ):
    """Run submodels for a single region

    Arguments
    ---------
    region : str
        Region name
    data : dict
        Data container

    Returns
    -------
    all_submodels : list
        List with submodel objects depending on submodels
    """
    submodel_names = assumptions.submodels_names

    # Get closest weather region object
    weather_region_id = weather_station_location.get_closest_station(
        latitude_reg=data['reg_coord'][region]['latitude'],
        longitude_reg=data['reg_coord'][region]['longitude'],
        weather_stations=weather_stations)

    # ----------------------------
    # Create Base year and current weather Regions
    # ----------------------------
    weather_region_cy = WeatherRegion(
        name=weather_region_id,
        latitude=weather_stations[weather_region_id]['latitude'],
        longitude=weather_stations[weather_region_id]['longitude'],
        assumptions=assumptions,
        technologies=assumptions.technologies,
        enduses=data['enduses'],
        temp_by=data['temp_data'][weather_by][weather_region_id],
        temp_cy=data['temp_data'][weather_yr][weather_region_id],
        tech_lp=data['tech_lp'],
        sectors=data['sectors'],
        crit_temp_min_max=criterias['crit_temp_min_max'])

    # Get region specific disaggregated fuel
    region_fuel_disagg = get_disaggregated_fuel_of_reg(
        submodel_names, data['fuel_disagg'], region)

    # ----------------------------
    # Create Region
    # ----------------------------
    region_obj = Region(
        name=region,
        longitude=data['reg_coord'][region]['longitude'],
        latitude=data['reg_coord'][region]['latitude'],
        region_fuel_disagg=region_fuel_disagg)

    for submodel_name in submodel_names:
        for sector in data['sectors'][submodel_name]:
            for enduse in data['enduses'][submodel_name]:

                # ------------------------------------------------------
                # Configure and select correct Enduse specific inputs
                # ------------------------------------------------------
                if submodel_name == 'industry' and enduse != "is_space_heating":
                    flat_profile_crit = True
                else:
                    flat_profile_crit = False

                if not data['dw_stocks'][submodel_name]:
                    dw_stock = False
                else:
                    dw_stock = data['dw_stocks'][submodel_name][region_obj.name]

                # ---------------
                # Create submodel for region and enduse
                # ---------------
                yield endusefunctions.Enduse(
                    submodel_name=submodel_name,
                    region=region_obj.name,
                    scenario_data=data['scenario_data'],
                    assumptions=assumptions,
                    load_profiles=weather_region_cy.load_profiles,
                    base_yr=assumptions.base_yr,
                    curr_yr=assumptions.curr_yr,
                    enduse=enduse,
                    sector=sector,
                    fuel=region_obj.fuels[submodel_name][enduse][sector],
                    tech_stock=weather_region_cy.tech_stock[submodel_name],
                    heating_factor_y=weather_region_cy.f_heat[submodel_name],
                    cooling_factor_y=weather_region_cy.f_colling[submodel_name],
                    fuel_tech_p_by=assumptions.fuel_tech_p_by[enduse][sector],
                    criterias=criterias,
                    strategy_vars=assumptions.regional_vars[region_obj.name],
                    fueltypes_nr=assumptions.fueltypes_nr,
                    fueltypes=assumptions.fueltypes,
                    dw_stock=dw_stock,
                    reg_scen_drivers=assumptions.scenario_drivers,
                    flat_profile_crit=flat_profile_crit)
    #return weather_region_id

def fuel_aggr(
        sector_models,
        sum_crit,
        attribute_non_technology,
        attribute_technologies,
        technologies,
        shape_aggregation_array
    ):
    """Collect hourly data from all regions and sum across
    all fuel types and enduses

    Arguments
    ----------
    sector_models : list[Enduse]
        Sector models
    sum_crit : str
        Criteria

    fueltypes_nr : int
        Number of fueltypes
    attribute_to_get : str
        Attribue to sumarise
    attribute_non_technology : str
        Attribute
    attribute_technologies : str
        Attribute 'techs_fuel_yh'
    technologies : dict
        Technologies
    shape_aggregation_array : array
        Input array to aggregate fuel

    Returns
    -------
    input_array : array
        Summarised array
    """
    input_array = np.zeros(shape_aggregation_array, dtype="float")

    for enduse_object in sector_models:

        fuels = get_fuels_yh(
            enduse_object,
            attribute_technologies)

        if isinstance(fuels, dict):
            for tech, fuel_tech in fuels.items():
                tech_fueltype = technologies[tech].fueltype_int
                input_array[tech_fueltype] += fuel_tech
        else:
            # Fuel per technology
            fuels = get_fuels_yh(
                enduse_object,
                attribute_non_technology)
            input_array += fuels

    if sum_crit == 'no_sum':
        return input_array
    elif sum_crit == 'sum':
        return np.sum(input_array)

def get_fuels_yh(enduse_object, attribute_to_get):
    """Get yh load profile and assign yh shape
    for enduses with flat load profiles

    Arguments
    ----------
    enduse_object : dict
        Object of submodel run
    attribute_to_get : str
        Attribute to read out

    Returns
    -------
    fuels : array
        Fuels with flat load profile

    Note
    -----
    For enduses where 'flat_profile_crit' in Enduse Class is True
    a flat load profile is generated. Otherwise, the yh as calculated
    for each enduse is used
    """
    if enduse_object.flat_profile_crit:

        # Annual fuel
        fuels_reg_y = enduse_object.fuel_y

        # Get flat load profile
        flat_shape_yd, flat_shape_yh, _ = generic_shapes.flat_shape()

        if attribute_to_get == 'shape_non_peak_y_dh':
            fuels = fuels_reg_y * flat_shape_yh
        elif attribute_to_get == 'shape_non_peak_yd':
            fuels = fuels_reg_y * flat_shape_yd
        elif attribute_to_get == 'fuel_yh' or attribute_to_get == 'techs_fuel_yh':
            f_hour = 1 / 8760
            flat_shape = np.full((enduse_object.fuel_y.shape[0], 365, 24), f_hour, dtype="float")
            fuels = fuels_reg_y[:, np.newaxis, np.newaxis] * flat_shape
    else: #If not flat shape, use yh load profile of enduse
        fuels = getattr(enduse_object, attribute_to_get)

    return fuels

def averaged_season_hourly(
        fueltype_region_yh,
        fueltypes,
        fueltypes_nr,
        seasons,
        reg_nrs
    ):
    """Calculate averaged hourly values for each season

    Arguments
    ---------
    averaged_h : dict
        Averaged hours per season (season, fueltype, array_nr_reg, 24)
    fueltype_region_yh : array
        Fuel of region (fueltype, yearday)
    reg_array_nr : int
        Integer of region
    fueltypes : dict
        Fueltype lookup
    ed_fueltype_regs_yh : array
       (fueltypes_nr, reg_nrs, yearhours_nrs)

    Return
    ------
    averaged_h : dict
        Averaged hourly value per season {season: array(fuetlype, region, 24)}
    """
    averaged_h = {
        'summer' : np.zeros((fueltypes_nr, reg_nrs, 24), dtype="float"),
        'spring': np.zeros((fueltypes_nr, reg_nrs, 24), dtype="float"),
        'winter': np.zeros((fueltypes_nr, reg_nrs, 24), dtype="float"),
        'autumn': np.zeros((fueltypes_nr, reg_nrs, 24), dtype="float")}

    fueltype_region_yh_365_24 = fueltype_region_yh.reshape(
        fueltype_region_yh.shape[0], fueltype_region_yh.shape[1], 365, 24)

    for fueltype in fueltypes:
        for season, yeardays_modelled in seasons.items():
            for region_nr in range(reg_nrs):
                for yearday in yeardays_modelled:
                    averaged_h[season][fueltype][region_nr] += fueltype_region_yh_365_24[fueltype][region_nr][yearday]

    # Calculate average hourly values for every season
    for season, yeardays_modelled in seasons.items():
        for fueltype in fueltypes:
            averaged_h[season][fueltype] = averaged_h[season][fueltype] / len(yeardays_modelled)

    return averaged_h

def create_virtual_dwelling_stocks(regions, curr_yr, data, criterias):
    """Create virtual dwelling stocks for residential
    and service sector.

    If no floor area is avilable, calculate average floor
    area with population information

    Arguments
    ---------

    """
    rs_dw_stock = defaultdict(dict)
    ss_dw_stock = defaultdict(dict)

    for region in regions:

        # -------------
        # Residential dwelling stocks
        # -------------
        dwtype = lookup_tables.basic_lookups()['dwtype']

        # Base year
        rs_dw_stock[region][data['assumptions'].base_yr] = dw_stock.rs_dw_stock(
            region,
            data['assumptions'],
            data['scenario_data'],
            data['assumptions'].sim_yrs,
            dwtype,
            data['enduses']['residential'],
            data['reg_coord'],
            data['assumptions'].scenario_drivers,
            data['assumptions'].base_yr,
            data['assumptions'].base_yr,
            criterias['virtual_building_stock_criteria'])

        # Current year
        rs_dw_stock[region][curr_yr] = dw_stock.rs_dw_stock(
            region,
            data['assumptions'],
            data['scenario_data'],
            data['assumptions'].sim_yrs,
            dwtype,
            data['enduses']['residential'],
            data['reg_coord'],
            data['assumptions'].scenario_drivers,
            curr_yr,
            data['assumptions'].base_yr,
            criterias['virtual_building_stock_criteria'])

        # -------------
        # Service dwelling stocks
        # -------------
        # base year
        ss_dw_stock[region][data['assumptions'].base_yr] = dw_stock.ss_dw_stock(
            region,
            data['enduses']['service'],
            data['sectors']['service'],
            data['scenario_data'],
            data['reg_coord'],
            data['assumptions'],
            data['assumptions'].base_yr,
            data['assumptions'].base_yr,
            criterias['virtual_building_stock_criteria'])

        # current year
        ss_dw_stock[region][curr_yr] = dw_stock.ss_dw_stock(
            region,
            data['enduses']['service'],
            data['sectors']['service'],
            data['scenario_data'],
            data['reg_coord'],
            data['assumptions'],
            curr_yr,
            data['assumptions'].base_yr,
            criterias['virtual_building_stock_criteria'])

    return dict(rs_dw_stock), dict(ss_dw_stock)

def create_dwelling_stock(regions, curr_yr, data):
    """Create dwelling stock based on NEWCASTLE data

    TODO: Implement

    Arguments
    ---------

    Returns
    -------
    """
    #data['rs_dw_stock'][region][curr_yr] = dw_stock.createNEWCASTLE_dwelling_stock(
    # self.curr_yr,
    # region,
    # )
    #data['ss_dw_stock'][region][curr_yr] = dw_stock.createNEWCASTLE_dwelling_stock(self.curr_yr)
    return data

def aggregate_result_unconstrained(
        nr_of_submodels,
        lookup_sector_enduses,
        ed_enduse_fueltype_regs_yh,
        fueltypes_nr,
        reg_nrs
    ):
    """Aggregated unconstrained results. Summarise energy demand
    of unconstrained mode (heat is provided)

    Arguments
    ---------
    nr_of_submodels : int
        Number of submodels
    submodels_enduses : dict
        Submodels and enduses
    ed_enduse_fueltype_regs_yh : array
        Fuel array
    fueltypes_nr : int
        Number of fueltypes
    reg_nrs : int
        Number of regions

    Returns
    --------
    constrained_array : array
        assumptions.nr_of_submodels, reg_nrs, fueltypes_nr, 8760
    """
    constrained_array = np.zeros((nr_of_submodels, reg_nrs, fueltypes_nr, 8760), dtype="float")

    for submodel_nr, enduse_array_nrs in lookup_sector_enduses.items():
        for enduse_array_nr in enduse_array_nrs:
            for fueltype_nr in range(fueltypes_nr):
                for region_nr in range(reg_nrs):
                    constrained_array[submodel_nr][region_nr][fueltype_nr] += ed_enduse_fueltype_regs_yh[enduse_array_nr][fueltype_nr][region_nr]

    return constrained_array

def aggregate_single_region(
        reg_nrs,
        lookup_enduses,
        aggr_results,
        reg_array_nr,
        all_submodels,
        mode_constrained,
        fueltypes_nr,
        enduse_space_heating,
        technologies
    ):
    """Aggregate results for a single region

    Parameters
    ----------
    aggr_results : dict
        Contains alls results to aggregate
        (key of this will be made self.attributes)
    reg_array_nr : int
        Region array number
    all_submodels : list[Enduse]
        Submodel objects
    mode_constrained : bool
        Mode of how to run the model
    fueltypes : dict
        Fueltypes lookup
    fueltypes_nr : int
        Number of fueltypes
    seasons : dict
        Seasons
    enduse_space_heating : list
        All heating enduses
    technologies : dict
        Technologies
    write_txt_additional_results : bool
        Criteria whether additional results are aggregated
        for plotting purposes going beyond the SMIF framework

    Returns
    --------
    aggr_results : dict
        Contains all aggregated results
    """

    # Get enduse number for enduse
    submodel_to_idx = {
        name: idx
        for idx, name in enumerate(lookup_tables.basic_lookups()['submodels_names'])
    }

    for enduse_object in all_submodels: # Iterate all simulation results
        # -----------------------------------------------------------------
        # Aggregate fuel of all technologies
        # -----------------------------------------------------------------
        techs_fueltypes_yh = get_fuels_yh(enduse_object, 'techs_fuel_yh')
        enduse_array_nr = lookup_enduses[enduse_object.enduse]
        submodel_nr = submodel_to_idx[enduse_object.submodel_name]

        if isinstance(techs_fueltypes_yh, dict):

            for tech, fuel_tech in techs_fueltypes_yh.items():
                tech_fueltype = technologies[tech].fueltype_int
                reshaped_fuel = fuel_tech.reshape(8760)

                # -------------------------------------------            
                # Add all technologies to main aggreator (heating & non heating)
                # -------------------------------------------
                aggr_results['ed_enduse_fueltype_regs_yh'][enduse_array_nr][tech_fueltype][reg_array_nr] += reshaped_fuel

                # -----------------------------------------------------------------
                # Aggregate fuel of constrained technologies for heating
                # Only add heating technologies to constrained results
                # -----------------------------------------------------------------
                if mode_constrained:
                    if enduse_object.enduse in enduse_space_heating:

                        # Aggregate Submodel (sector) specific enduse for fueltype
                        if tech in aggr_results['results_constrained'].keys():
                            aggr_results['results_constrained'][tech][submodel_nr][reg_array_nr][tech_fueltype] += reshaped_fuel
                        else:
                            if tech not in aggr_results['results_constrained']:
                                aggr_results['results_constrained'][tech] = np.zeros((len(submodel_to_idx), reg_nrs, fueltypes_nr, 8760), dtype="float")
                            aggr_results['results_constrained'][tech][submodel_nr][reg_array_nr][tech_fueltype] += reshaped_fuel

                    else:
                        aggr_results['results_unconstrained_no_heating'][submodel_nr][reg_array_nr][tech_fueltype] += reshaped_fuel
                else:
                    aggr_results['results_unconstrained_no_heating'][submodel_nr][reg_array_nr][tech_fueltype] += reshaped_fuel

        else:
            # Only add heating technologies to constrained results
            fueltype_yh_365_24 = get_fuels_yh(enduse_object, 'fuel_yh')
            fueltype_yh_8760 = fueltype_yh_365_24.reshape(fueltypes_nr, 8760)

            for fueltype_nr, fuels_8760 in enumerate(fueltype_yh_8760):
                aggr_results['ed_enduse_fueltype_regs_yh'][enduse_array_nr][fueltype_nr][reg_array_nr] += fuels_8760
                
                aggr_results['results_unconstrained_no_heating'][submodel_nr][reg_array_nr][fueltype_nr] += fuels_8760

    return aggr_results

def initialise_result_container(
        fueltypes_nr,
        reg_nrs,
        lookup_enduses,
        nr_of_submodels
    ):
    """Create container with empty dict or arrays
    as values in a dict. This is used to aggregate the
    region calculation results

    Arguments
    ---------
    fueltypes_nr : int
        Number of fueltypes
    sectors : list
        Sectors
    lookup_enduses : dict
        Lookup of enduses of array position
    reg_nrs : int
        Number of regions

    Returns
    -------
    container : dict
        Contained with all empty correctly formated values for aggregation
    """
    container = {}
    container['results_constrained'] = {}
    container['results_unconstrained_no_heating'] = np.zeros((
        nr_of_submodels, reg_nrs, fueltypes_nr, 8760), dtype="float")
    container['ed_enduse_fueltype_regs_yh'] = np.zeros((
        len(lookup_enduses), fueltypes_nr, reg_nrs, 8760), dtype="float")

    return container

def get_enduse_from_array_nr(enduse_array_yr_to_get, lookup_enduse):
    """Get enduse string of enduse array number

    Arguments
    ---------
    enduse_array_yr_to_get : int
        Array positoin
    lookup_enduse : dict
        dictionary with enduse and enduse nr

    Returns
    -------
    enduse_str_to_get : str
        Enduse string name
    """
    for enduse_str, enduse_arry_nr in lookup_enduse.items():
        if enduse_arry_nr == enduse_array_yr_to_get:
            enduse_str_to_get = enduse_str
            break

    return enduse_str_to_get

def aggregate_from_full_results(
        lookup_enduse,
        aggregated_container,
        full_sim_data,
        time_resolution,
        per_region,
        per_sector,
        per_enduse
    ):
    """Function to read from full simulation data specific demands

    Argument
    --------
    lookup_enduse : dict
        Enduse lookup
    aggregated_container : dict
        Container to aggregate
    full_sim_data : array
         Modelling results per submodel, enduse, region, fueltype, 8760h
    time_resolution : str
        Either 'annual', 'hourly', '360_24_h', '8760_h'
    """
    for enduse_array_yr, reg_fueltype_8760h in enumerate(full_sim_data):

        # Test if reshaping and annual or hourly
        if time_resolution == 'annual':
            demand = np.sum(reg_fueltype_8760h, axis=2) # Sum all hours
        elif time_resolution == '8760_h':
            demand = reg_fueltype_8760h
        elif time_resolution == '365_24':
            demand = reg_fueltype_8760h.reshape(
                reg_fueltype_8760h.shape[0],
                reg_fueltype_8760h.shape[1],
                365, 24)
        else:
            raise Exception("Provide either 'annual' or 'hourly' or '365_24'")

        # Aggregate according to criteria
        if per_sector:
            if per_region:
                if per_enduse:
                    pass
                else:
                    pass
            else:
                if per_enduse:
                    pass
                else:
                    pass
        else:
            if per_region:
                if per_enduse:
                    aggregated_container[enduse_array_yr] += demand
                else:
                    aggregated_container += demand
            else:
                # Sum across all regions
                demand = np.sum(demand, axis=1)

                # Get enduse str
                enduse_str = get_enduse_from_array_nr(enduse_array_yr, lookup_enduse)

                if per_enduse:
                    aggregated_container[enduse_str] += demand
                else:
                    aggregated_container += demand

    return aggregated_container



"""Contains the `Enduse` Class. This is the most important class
where the change in enduse specific energy demand is simulated
depending on scenaric assumptions"""
import logging
import math
import copy
import numpy as np

from energy_demand.basic import testing_functions
from energy_demand.profiles import load_factors as lf
from energy_demand.technologies import diffusion_technologies
from energy_demand.technologies import fuel_service_switch
from energy_demand.technologies import tech_related
from energy_demand.basic import basic_functions
from energy_demand.basic import lookup_tables

class Enduse(object):
    """Enduse Class for all endueses in each SubModel

    For every region and sector, a different instance
    is generated. In this class, first the change in
    energy demand is calculated on a annual temporal scale.
    Calculations are performed in a cascade (e.g. first
    reducing climate change induced savings, then substracting
    further behavioral savings etc.). After annual calculations,
    the demand is converted to hourly demand.

    Also within this function, the fuel inputs are converted
    to energy service (short: service) and converted back to
    fuels (e.g. electricit).

    Arguments
    ----------
    region : str
        Region name
    scenario_data : dict
        Scenario data
    assumptions : dict
        Assumptions
    load_profiles : dict
        Load profile stock
    base_yr : int
        Base year
    curr_yr : int
        Current year
    enduse : str
        Enduse name
    sector : str
        Sector name
    fuel : array
        Yearly fuel data for different fueltypes
    tech_stock : object
        Technology stock of region
    heating_factor_y : array
        Distribution of fuel within year to days (yd) (directly correlates with HDD)
    cooling_factor_y : array
        Distribution of fuel within year to days (yd) (directly correlates with CDD)
    fuel_tech_p_by : dict
        Fuel tech assumtions in base year
    regional_lp_stock : object
        Load profile stock
    dw_stock : object,default=False
        Dwelling stock
    reg_scen_drivers : bool,default=None
        Scenario drivers per enduse
    flat_profile_crit : bool,default=False
        Criteria of enduse has a flat shape or not
    make_all_flat : bool
        Crit to make everything flat

    Note
    ----
    - Load profiles are assigned independently of the fueltype, i.e.
      the same profiles are assumed to hold true across different fueltypes

    - ``self.fuel_y`` is always overwritten
      in the cascade of calculations

    Warning
    -------
    Not all enduses have technologies assigned. Load peaks are derived
    from techstock in case technologies are defined. Otherwise enduse load
    profiles are used.
    """
    def __init__(
            self,
            submodel_name,
            region,
            scenario_data,
            assumptions,
            load_profiles,
            base_yr,
            curr_yr,
            enduse,
            sector,
            fuel,
            tech_stock,
            heating_factor_y,
            cooling_factor_y,
            fuel_tech_p_by,
            criterias,
            strategy_vars,
            fueltypes_nr,
            fueltypes,
            dw_stock=False,
            reg_scen_drivers=None,
            flat_profile_crit=False,
            make_all_flat=False
        ):
        """Enduse class constructor
        """
        self.submodel_name = submodel_name
        self.region = region
        self.enduse = enduse
        self.fuel_y = fuel
        self.flat_profile_crit = flat_profile_crit
        self.techs_fuel_yh = None

        if np.sum(fuel) == 0: #If enduse has no fuel return empty shapes
            self.flat_profile_crit = True
            self.fuel_y = fuel
            self.fuel_yh = 0
            self.enduse_techs = []
        else:
            #logging.debug("------INFO  {} {} {}  {}".format(self.enduse, sector, region, curr_yr))
            self.enduse_techs = get_enduse_techs(fuel_tech_p_by)

            # -----------------------------
            # Cascade of annual calculations
            # -----------------------------
            _fuel_new_y = apply_weather_correction(
                enduse,
                self.fuel_y,
                cooling_factor_y,
                heating_factor_y,
                assumptions.enduse_space_heating,
                assumptions.ss_enduse_space_cooling)
            self.fuel_y = _fuel_new_y

            _fuel_new_y = apply_smart_metering(
                enduse,
                self.fuel_y,
                assumptions.smart_meter_assump,
                strategy_vars,
                curr_yr)
            self.fuel_y = _fuel_new_y

            _fuel_new_y = generic_demand_change(
                enduse,
                sector,
                self.fuel_y,
                strategy_vars['generic_enduse_change'],
                curr_yr)
            self.fuel_y = _fuel_new_y

            _fuel_new_y = apply_scenario_drivers(
                enduse=enduse,
                sector=sector,
                fuel_y=self.fuel_y,
                dw_stock=dw_stock,
                region=region,
                gva_industry=scenario_data['gva_industry'],
                gva_per_head=scenario_data['gva_per_head'],
                population=scenario_data['population'],
                reg_scen_drivers=reg_scen_drivers,
                base_yr=base_yr,
                curr_yr=curr_yr)
            self.fuel_y = _fuel_new_y

            # Apply cooling scenario variable
            _fuel_new_y = apply_cooling(
                enduse,
                self.fuel_y,
                strategy_vars['cooled_floorarea'],
                assumptions.cooled_ss_floorarea_by,
                curr_yr)
            self.fuel_y = _fuel_new_y

            # Industry related change
            _fuel_new_y = industry_enduse_changes(
                enduse,
                sector,
                curr_yr,
                strategy_vars,
                self.fuel_y,
                assumptions)
            self.fuel_y = _fuel_new_y

            # Generic fuel switch of an enduse and sector
            _fuel_new_y = generic_fuel_switch(
                enduse,
                sector,
                curr_yr,
                base_yr,
                strategy_vars['generic_fuel_switch'],
                self.fuel_y)
            self.fuel_y = _fuel_new_y

            # ----------------------------------
            # Hourly Disaggregation
            # ----------------------------------
            if self.enduse_techs == []:
                # If no technologies are defined for an enduse, the load profiles
                # are read from dummy shape, which show the load profiles of the whole enduse.
                # No switches can be implemented and only overall change of enduse
                if flat_profile_crit:
                    pass
                else:
                    fuel_yh = assign_lp_no_techs(
                        enduse,
                        sector,
                        load_profiles,
                        self.fuel_y,
                        make_all_flat=make_all_flat)

                    # Demand management for non-technology enduse
                    self.fuel_yh = load_shifting(
                        enduse,
                        fuel_yh,
                        mode_constrained=False,
                        param_lf_improved_cy=strategy_vars['dm_improvement'][enduse][curr_yr],
                        make_all_flat=make_all_flat)

                    ##assert not testing_functions.test_if_minus_value_in_array(self.fuel_yh)

            else: #If technologies are defined for an enduse
                mode_constrained = get_enduse_configuration(
                    criterias['mode_constrained'],
                    enduse,
                    assumptions.enduse_space_heating)

                # ------------------------------------
                # Calculate regional energy service
                # ------------------------------------
                s_tot_y_cy, s_tech_y_by = fuel_to_service(
                    enduse,
                    sector,
                    self.fuel_y,
                    fuel_tech_p_by,
                    tech_stock,
                    fueltypes,
                    mode_constrained)

                # ------------------------------------
                # Reduction of service because of heat recovery
                # ------------------------------------
                s_tot_y_cy, s_tech_y_cy = apply_heat_recovery(
                    enduse,
                    strategy_vars['heat_recovered'],
                    s_tot_y_cy,
                    s_tech_y_by,
                    curr_yr)

                # ------------------------------------
                # Reduction of service because of improvement in air leakeage
                # ------------------------------------
                s_tot_y_cy, s_tech_y_cy = apply_air_leakage(
                    enduse,
                    strategy_vars['air_leakage'],
                    s_tot_y_cy,
                    s_tech_y_cy,
                    curr_yr)

                # --------------------------------
                # Switches
                # --------------------------------
                s_tech_y_cy = apply_service_switch(
                    enduse=enduse,
                    s_tech_y_cy=s_tech_y_cy,
                    all_technologies=self.enduse_techs,
                    curr_yr=curr_yr,
                    base_yr=base_yr,
                    sector=sector,
                    annual_tech_diff_params=strategy_vars['annual_tech_diff_params'][enduse][sector],
                    crit_switch_happening=assumptions.crit_switch_happening)

                # -------------------------------------------
                # Convert annual service to fuel per fueltype
                # -------------------------------------------
                self.fuel_y, fuel_tech_y = service_to_fuel(
                    enduse,
                    sector,
                    s_tech_y_cy,
                    tech_stock,
                    fueltypes_nr,
                    fueltypes,
                    mode_constrained)

                # Delete all technologies with no fuel assigned
                # This would speed up the calculations and use less memory
                '''for tech, fuel_tech in fuel_tech_y.items():
                    if np.sum(fuel_tech) == 0:
                        self.enduse_techs.remove(tech)'''
                
                ##for i in fuel_tech_y.values():
                ##    assert not testing_functions.test_if_minus_value_in_array(i)
                ##assert not testing_functions.test_if_minus_value_in_array(self.fuel_y)

                # ------------------------------------------
                # Assign load profiles
                # ------------------------------------------
                if self.flat_profile_crit:
                    pass
                else:
                    fuel_yh = calc_fuel_tech_yh(
                        enduse,
                        sector,
                        self.enduse_techs,
                        fuel_tech_y,
                        load_profiles,
                        fueltypes_nr,
                        fueltypes,
                        mode_constrained)

                    # --------------------------------------
                    # Peak shifting
                    # --------------------------------------
                    if mode_constrained:
                        self.fuel_yh = None

                        # If no demand management improvenent, no peak shifting
                        if strategy_vars['dm_improvement'][enduse][curr_yr] == 0:
                            self.techs_fuel_yh = fuel_yh
                        else:
                            self.techs_fuel_yh = load_shifting_multiple_tech(
                                enduse,
                                fueltypes,
                                self.enduse_techs,
                                assumptions.technologies,
                                fuel_yh,
                                param_lf_improved_cy=strategy_vars['dm_improvement'][enduse][curr_yr])

                            ##assert not testing_functions.test_if_minus_value_in_array(self.fuel_y)
                            ##for key, i in self.techs_fuel_yh.items():
                            ##    assert not testing_functions.test_if_minus_value_in_array(i)
                    else:
                        self.fuel_yh = load_shifting(
                            enduse,
                            fuel_yh,
                            mode_constrained=False,
                            param_lf_improved_cy=strategy_vars['dm_improvement'][enduse][curr_yr],
                            make_all_flat=make_all_flat)

                        ##if testing_functions.test_if_minus_value_in_array(self.fuel_yh):
                        ##    raise Exception("Minus fuel value detected: {}  {} {}".format(enduse, sector, np.sum(self.fuel_yh)))

def load_shifting_multiple_tech(
        enduse,
        fueltypes,
        enduse_techs,
        technologies,
        fuel_yh,
        param_lf_improved_cy
    ):
    """Shift demand in case of multiple technologiesself.

    Check how much of each technology is shifted in peak hours.
    Calculate the absolute and relative demand of shifited demand
    for every technology. Then substract and assign the new demands
    per technology.
    """
    fuel_tech_shifted = {}

    for fueltype in fueltypes:

        # Get technologies of fueltype
        technologies_fueltype = tech_related.get_technologies_per_fueltype(
            enduse_techs, technologies, fueltype)

        if technologies_fueltype != []:

            # Sum across al technologies
            sum_all_techs = np.zeros((365, 24))
            for tech in technologies_fueltype:
                sum_all_techs += fuel_yh[tech]

            # Shift demand for enduse (summed demand of technologies)
            sum_all_techs_shifted = load_shifting(
                enduse,
                sum_all_techs,
                mode_constrained=True,
                param_lf_improved_cy=param_lf_improved_cy,
                make_all_flat=False)

            # Shift demand of peak (removed demand)
            removed_demand = sum_all_techs - sum_all_techs_shifted
            removed_demand = removed_demand.reshape(8760)
            removed_demand[removed_demand < 0] = 0
            tot_removed_demand = sum(removed_demand)
            shape_removed = (1 / tot_removed_demand) * removed_demand
            shape_removed[np.isnan(shape_removed)] = 0
            peak_hour_positions = list(np.where(removed_demand > 0)[0])

            # Demand in off-peak (demand added)
            added_demand = sum_all_techs - sum_all_techs_shifted
            added_demand = added_demand.reshape(8760)
            added_demand[added_demand > 0] = 0
            added_demand = abs(added_demand)
            tot_added_demand = sum(added_demand)
            shape_added = (1 / tot_added_demand) * added_demand
            shape_added[np.isnan(shape_added)] = 0

            assert round(np.sum(tot_removed_demand), 6) == round(tot_added_demand, 6)

            for tech in technologies_fueltype:

                fuel_tech_8760 = copy.copy(fuel_yh[tech].reshape(8760)) 
                sum_all_techs_8760 = sum_all_techs.reshape(8760)

                # Calculate share of technology specific demand to overall enduse demand per hour before shift
                h_contribution_before = (1 / sum_all_techs_8760) * fuel_tech_8760
                h_contribution_before[np.isnan(h_contribution_before)] = 0

                # Get absolute demand in peak hours
                tech_peak_shifted_demand = 0
                for peak_hour_pos in peak_hour_positions:
                    tech_peak_shifted_demand += h_contribution_before[peak_hour_pos] * removed_demand[peak_hour_pos]

                # Percentage of total shifted demand of technology
                p_abs_tech_shifted = (1.0 / tot_removed_demand) * tech_peak_shifted_demand

                if np.isnan(p_abs_tech_shifted):
                    p_abs_tech_shifted = 0

                # Add shifted demand to total demand
                fuel_tech_8760 += (p_abs_tech_shifted * tot_removed_demand) * shape_added

                # Substract peak from total demand
                fuel_tech_8760 -= (p_abs_tech_shifted * tot_removed_demand) * shape_removed

                assert round(np.sum(fuel_tech_8760), 6) == round(np.sum(fuel_yh[tech]), 6)

                fuel_tech_shifted[tech] = fuel_tech_8760.reshape(365, 24)

    return fuel_tech_shifted

def load_shifting(
        enduse,
        fuel_yh,
        mode_constrained,
        param_lf_improved_cy,
        make_all_flat=False
    ):
    """This function shifts peak per of this enduse
    depending on peak shifting factors.
    So far only inter day load shifting

    Arguments
    ----------

    fuel_yh : array
        Fuel per hours
    mode_constrained : bool
        Running mode, If mode_constrained, always only one technology imported
    param_lf_improved_cy : dict
        Assumptions of strategy variables (dm_improvement)
        Load factor improvement parameter in current year
    make_all_flat : bool
        If true, all shapes are flat

    Returns
    -------
    fuel_yh : array
        Fuel of yh
    """
    if param_lf_improved_cy == 0:
        pass # no load management
    else:

        # Calculate average for every day
        if mode_constrained:
            average_fuel_yd = np.average(fuel_yh, axis=1)
        else:
            average_fuel_yd = np.average(fuel_yh, axis=2)

        # Calculate load factors (only inter_day load shifting as for now)
        loadfactor_yd_cy = lf.calc_lf_d(
            fuel_yh, average_fuel_yd, mode_constrained)

        # Calculate current year load factors
        lf_improved_cy = calc_lf_improvement(
            param_lf_improved_cy,
            loadfactor_yd_cy)

        fuel_yh = lf.peak_shaving_max_min(
            lf_improved_cy,
            average_fuel_yd,
            fuel_yh,
            mode_constrained)

    # -------------------------------------------------
    # Convert all load profiles into flat load profiles
    # -------------------------------------------------
    if make_all_flat:
        if mode_constrained:
            sum_fueltypes_days = np.sum(fuel_yh)                 #sum over all hours
            average_fueltype = sum_fueltypes_days / 8760         # Average
            fuel_yh_empty = np.ones((fuel_yh.shape))
            fuel_yh = fuel_yh_empty * average_fueltype
        else:
            sum_fueltypes_days_h = np.sum(fuel_yh, 2)            #sum over all hours
            sum_fueltypes_days = np.sum(sum_fueltypes_days_h, 1) #sum over all days
            average_fueltype = sum_fueltypes_days / 8760         #Average per fueltype
            fuel_yh_empty = np.ones((fuel_yh.shape))
            fuel_yh = fuel_yh_empty * average_fueltype[:, np.newaxis, np.newaxis]

    return fuel_yh

def calc_lf_improvement(
        param_lf_improved_cy,
        loadfactor_yd_cy
    ):
    """Calculate load factor improvement

    Arguments
    ---------
    lf_improvement_ey : dict
        Load factor improvement until end year provided as decimal (1 == 100%)
    loadfactor_yd_cy : float
        Yd Load factor of current year

    Returns
    -------
    lf_improved_cy : str
        Improved load factor of current year
    peak_shift_crit : bool
        True: Peak is shifted, False: Peak isn't shifed
    """
    # Add load factor improvement to current year load factor
    lf_improved_cy = loadfactor_yd_cy + param_lf_improved_cy

    # Where load factor larger than zero, set to 1
    lf_improved_cy[lf_improved_cy > 1] = 1

    return lf_improved_cy

def assign_lp_no_techs(
        enduse,
        sector,
        load_profiles,
        fuel_y,
        make_all_flat
    ):
    """Assign load profiles for an enduse which has no technologies defined

    Arguments
    ---------
    enduse : str
        Enduse
    sector : str
        Enduse
    load_profiles : obj
        Load profiles
    fuel_y : array
        Fuels

    Returns
    -------
    fuel_yh : array (fueltype, 365, 24)
        Fuel yh
    """
    # Load profile for all fueltypes
    load_profile = load_profiles.get_lp(
        enduse, sector, 'placeholder_tech', 'shape_yh')

    fuel_yh = load_profile[:np.newaxis] * fuel_y[:, np.newaxis, np.newaxis]

    # Convert all load profiles into flat load profiles
    if make_all_flat:
        sum_fueltypes_days_h = np.sum(fuel_yh, 2)            #sum over all hours
        sum_fueltypes_days = np.sum(sum_fueltypes_days_h, 1) #sum over all days
        average_fueltype = sum_fueltypes_days / 8760         #Average per fueltype
        fuel_yh_empty = np.ones((fuel_yh.shape))
        fuel_yh = fuel_yh_empty * average_fueltype[:, np.newaxis, np.newaxis]

    return fuel_yh

def get_running_mode(enduse, mode_constrained, enduse_space_heating):
    """Checks which mode needs to be run for an enduse.

    Arguments
    -----------
    mode_constrained : bool
        Criteria of running mode
    enduse_space_heating : dict
        All heating enduses across all models

    Returns
    -------
    bool : bool
        The return value

    Note
    ----
    If 'crit_mode' == True, then overall heat is provided to
    the supply model not specified for technologies. Otherwise,
    heat demand is supplied per technology
    """
    if mode_constrained:
        return True
    elif not mode_constrained and enduse in enduse_space_heating:
        return False
    elif not mode_constrained and enduse not in enduse_space_heating:
        # All other not constrained enduses where technologies are defined
        # are run in 'constrained' mode (e.g. lighting)
        return True

def get_enduse_configuration(
        mode_constrained,
        enduse,
        enduse_space_heating,
    ):
    """Get enduse specific configuration

    Arguments
    ---------
    mode_constrained : bool
        Constrained mode criteria
    enduse : str
        Enduse
    enduse_space_heating : list
        All endueses classified as space heating
    base_yr, curr_yr : int
        Base, current, year
    """
    mode_constrained = get_running_mode(
        enduse,
        mode_constrained,
        enduse_space_heating)

    return mode_constrained

def get_peak_day_all_fueltypes(fuel_yh):
    """Iterate yh and get day containing the hour
    with the largest demand (across all fueltypes).

    Arguments
    ---------
    fuel_yh : array (fueltype, 365, 24)
        Fuel for every yh (fueltypes, yh)

    Return
    ------
    peak_day_nr : int
        Day with most fuel or service across all fueltypes
    """
    fuel_8760 = fuel_yh.reshape(fuel_yh.shape[0], 8760)

    # Sum all fuel across all fueltypes for every hour in a year
    all_fueltypes_tot_h = np.sum(fuel_8760, axis=0)

    if np.sum(all_fueltypes_tot_h) == 0:
        logging.warning("No peak can be found because no fuel assigned")
        return 0
    else:

        # Get day with maximum hour
        peak_day_nr = basic_functions.round_down(np.argmax(all_fueltypes_tot_h) / 24, 1)

        return int(peak_day_nr)

def get_peak_day(fuel_yh):
    """Iterate an array with entries and get
    entry nr with hightest value

    Arguments
    ---------
    fuel_yh : array (hours)
        Fuel for every day

    Return
    ------
    peak_day_nr : int
        Day with most fuel or service
    """
    if np.sum(fuel_yh) == 0:
        #logging.info("No peak can be found because no fuel assigned")
        # Return first entry of element (which is zero)
        return 0
    else:
        # Sum fuel within every hour for every day and get day with maximum fuel
        peak_day_nr = np.argmax(fuel_yh)

        return int(peak_day_nr)

def get_peak_day_single_fueltype(fuel_yh):
    """Iterate yh and get day with highes fuel for a single fueltype
    The day with most fuel is considered to
    be the peak day. Over the simulation period,
    the peak day may change date in a year.

    Arguments
    ---------
    fuel_yh : array (365, 24) or array (8760)
        Fuel for every yh (yh)

    Return
    ------
    peak_day_nr : int
        Day with most fuel or service
    peak_h : float
        Peak hour value
    """
    fuel_yh_8760 = fuel_yh.reshape(8760)

    if np.sum(fuel_yh_8760) == 0:
        #logging.info("No peak can be found because no fuel assigned")
        # Return first entry of element (which is zero)
        return 0, 0
    else:
        # Sum fuel within every hour for every day and get day with maximum fuel
        peak_day_nr = int(basic_functions.round_down(np.argmax(fuel_yh_8760) / 24, 1))

        peak_h = np.max(fuel_yh_8760)

        return int(peak_day_nr), peak_h

def get_trough_day_single_fueltype(fuel_yh):
    """Iterate yh and get day with lowest fuel for a single fueltype
    The day with most fuel is considered to
    be the peak day. Over the simulation period,
    the peak day may change date in a year.

    Arguments
    ---------
    fuel_yh : array (365, 24) or array (8760)
        Fuel for every yh (yh)

    Return
    ------
    trough_day_nr : int
        Day with most fuel or service
    trough_h : float
        Peak hour value
    """
    fuel_yh_8760 = fuel_yh.reshape(8760)

    if np.sum(fuel_yh_8760) == 0:
        #logging.info("No peak can be found because no fuel assigned")
        # Return first entry of element (which is zero)
        return 0, 0
    else:
        # Sum fuel within every hour for every day and get day with maximum fuel
        trough_day_nr = int(basic_functions.round_down(np.argmin(fuel_yh_8760) / 24, 1))

        fuel_yh_365_24 = fuel_yh_8760.reshape(365, 24)
        trough_peak_h = np.max(fuel_yh_365_24[trough_day_nr])

        return int(trough_day_nr), trough_peak_h

def get_enduse_techs(fuel_tech_p_by):
    """Get all defined technologies of an enduse

    Arguments
    ----------
    fuel_tech_p_by : dict
        Percentage of fuel per enduse per technology

    Return
    ------
    enduse_techs : list
        All technologies

    Note
    ----
    All technologies are read out, including those which
    are potentially defined in fuel or service switches.

    If for an enduse a dummy technology is defined,
    the technologies of an enduse are set to an empty
    list.

    Warning
    -------
    For every enduse technologes must either be defined
    for no fueltype or for all fueltypes
    """
    enduse_techs = []

    for tech_fueltype in fuel_tech_p_by.values():
        if 'placeholder_tech' in tech_fueltype.keys():
            return []
        else:
            enduse_techs += tech_fueltype.keys()

    return list(set(enduse_techs))

def calc_fuel_tech_yh(
        enduse,
        sector,
        enduse_techs,
        fuel_tech_y,
        load_profiles,
        fueltypes_nr,
        fueltypes,
        mode_constrained
    ):
    """Iterate fuels for each technology and assign shape yd and yh shape

    Arguments
    ----------
    fuel_tech_y : dict
        Fuel per technology in enduse
    tech_stock : object
        Technologies
    load_profiles : object
        Load profiles
    fueltypes_nr : dict
        Nr of fueltypes
    fueltypes : dict
        Fueltypes lookup
    mode_constrained : bool
        Mode criteria

    Return
    ------
    fuels_yh : array
        Fueltype storing hourly fuel for every fueltype (fueltype, 365, 24)
    """
    if mode_constrained:
        fuels_yh = {}
        for tech in enduse_techs:

            load_profile = load_profiles.get_lp(
                enduse, sector, tech, 'shape_yh')

            fuels_yh[tech] = fuel_tech_y[tech] * load_profile
    else:
        # --
        # Unconstrained mode, i.e. not technolog specific.
        # Store according to fueltype and heat
        # --
        fuels_yh = np.zeros((fueltypes_nr, 365, 24), dtype="float")

        for tech in enduse_techs:

            load_profile = load_profiles.get_lp(
                enduse, sector, tech, 'shape_yh')

            # If no fuel for this tech and not defined in enduse
            fuel_tech_yh = fuel_tech_y[tech] * load_profile

            fuels_yh[fueltypes['heat']] += fuel_tech_yh

    return fuels_yh

def service_to_fuel(
        enduse,
        sector,
        service_tech,
        tech_stock,
        fueltypes_nr,
        fueltypes,
        mode_constrained
    ):
    """Convert yearly energy service to yearly fuel demand.
    For every technology the service is taken and converted
    to fuel based on efficiency of current year

    Arguments
    ------
    enduse : str
        Enduse
    service_tech : dict
        Service per fueltype and technology
    tech_stock : object
        Technological stock
    fueltypes_nr : int
        Number of fueltypes
    fueltypes : dict
        Fueltypes
    mode_constrained : bool
        Mode running criteria

    Returns
    -------
    fuel_y : array
        Fuel per fueltype
    fuel_per_tech : dict
        Fuel per technology

    Note
    -----
        - Fuel = Energy service / efficiency
    """
    fuel_tech_y = {}
    fuel_y = np.zeros((fueltypes_nr), dtype="float")

    if mode_constrained:
        for tech, service in service_tech.items():
            tech_eff = tech_stock.get_tech_attr(
                enduse, sector, tech, 'eff_cy')
            fueltype_int = tech_stock.get_tech_attr(
                enduse, sector, tech, 'fueltype_int')

            fuel_tech = service / tech_eff # Convert to fuel

            fuel_tech_y[tech] = fuel_tech # Add fuel
            fuel_y[fueltype_int] += fuel_tech
    else:
        for tech, fuel_tech in service_tech.items():
            fuel_y[fueltypes['heat']] += fuel_tech
            fuel_tech_y[tech] = fuel_tech

    return fuel_y, fuel_tech_y

def fuel_to_service(
        enduse,
        sector,
        fuel_y,
        fuel_tech_p_by,
        tech_stock,
        fueltypes,
        mode_constrained
    ):
    """Converts fuel to energy service. Calculate energy service
    of each technology based on assumptions about base year fuel
    shares of an enduse (`fuel_tech_p_by`).

    Arguments
    ----------
    enduse : str
        Enduse
    fuel_y : array
        Fuel per fueltype
    fuel_tech_p_by : dict
        Fuel composition of base year for every fueltype for each
        enduse (assumtions for national scale)
    tech_stock : object
        Technology stock of region
    fueltypes : dict
        Fueltype look-up
    mode_constrained : bool
        Criteria about mode

    Return
    ------
    tot_s_y : array
        Total annual energy service per technology
    s_tech_y : dict
        Total annual energy service per technology

    Note
    -----
    -   Efficiency changes of technologis are considered.
    -   Energy service = fuel * efficiency
    -   This function can be run in two modes, depending on `mode_constrained`
    -   The base year efficiency is taken because the actual service can
        only be calculated with base year.
        Efficiencies are only considered if converting back to fuel
        The 'self.fuel_y' is taken because the actual
        service was reduced e.g. due to smart meters or temperatur changes
    """
    s_tech_y = {}
    s_tot_y = 0

    # Calculate share of service
    for fueltype_int, tech_list in fuel_tech_p_by.items():

        # Get technologies to iterate
        if tech_list == {} and fuel_y[fueltype_int] == 0:   # No technology or fuel defined
            techs_with_fuel = {}
        elif tech_list == {} and fuel_y[fueltype_int] > 0:  # Fuel defined but no technologies
            fueltype_str = tech_related.get_fueltype_str(fueltypes, fueltype_int)
            placeholder_tech = 'placeholder_tech__{}'.format(fueltype_str)
            techs_with_fuel = {placeholder_tech: 1.0}
        else:
            techs_with_fuel = tech_list

        for tech, fuel_share in techs_with_fuel.items():

            #Constrained version
            if mode_constrained:
                tech_eff = tech_stock.get_tech_attr(enduse, sector, tech, 'eff_by')

                # Get fuel share and convert fuel to service per technology
                s_tech = fuel_y[fueltype_int] * fuel_share * tech_eff
                s_tech_y[tech] = s_tech
                s_tot_y += s_tech # Sum total yearly service
            else:
                #Unconstrained version (efficiencies are not considered, because not technology specific service calculation)
                fuel_tech = fuel_y[fueltype_int] * fuel_share # Calculate fuel share
                s_tech_y[tech] = fuel_tech
                s_tot_y += fuel_tech # Sum total yearly service

    return s_tot_y, s_tech_y

def apply_heat_recovery(
        enduse,
        heat_recovered,
        service,
        service_techs,
        curr_yr
    ):
    """Reduce heating demand according to assumption on heat reuse

    Arguments
    ----------
    enduse : str
        Enduse
    strategy_vars : dict
        Strategy variables
    service : dict or array
        Service of current year
    crit_dict : str
        Criteria to run function differently
    curr_yr : int
        Current year

    Returns
    -------
    service_reduced : dict or array
        Reduced service after assumption on reuse

    Note
    ----
    A standard sigmoid diffusion is assumed from base year to end year
    """
    try:
        # Fraction of heat recovered in current year
        heat_recovered_p_cy = heat_recovered[enduse][curr_yr]

        if heat_recovered_p_cy == 0:
            return service, service_techs
        else:
            # Apply to technologies each stored in dictionary
            service_reduced_techs = {}
            for tech, service_tech in service_techs.items():
                service_reduced_techs[tech] = service_tech * (1.0 - heat_recovered_p_cy)

            # Apply to array
            service_reduced = service * (1.0 - heat_recovered_p_cy)

            return service_reduced, service_reduced_techs

    except KeyError:

        # no recycling defined
        return service, service_techs

def apply_air_leakage(
        enduse,
        air_leakage,
        service,
        service_techs,
        curr_yr
    ):
    """Reduce heating demand according to assumption on
    improvements in air leaking

    Arguments
    ----------
    enduse : str
        Enduse
    air_leakage : dict
        Strategy variables
    service : dict or array
        Service of current year
    crit_dict : str
        Criteria to run function differently
    curr_yr : int
        Current year

    Returns
    -------
    service_reduced : dict or array
        Service after assumptions on air leaking improvements

    Note
    ----
    A standard sigmoid diffusion is assumed from base year to end year
    """
    try:
        # Fraction of heat recovered in current year
        air_leakage_improvement_cy = air_leakage[enduse][curr_yr]

        if air_leakage_improvement_cy == 0:
            return service, service_techs
        else:
            air_leakage_by = 1
            air_leakage_cy = 1 - air_leakage_improvement_cy

            f_improvement = air_leakage_cy / air_leakage_by

            # Apply to technologies each stored in dictionary or array
            service_reduced_techs = {}
            for tech, service_tech in service_techs.items():
                service_reduced_techs[tech] = service_tech * f_improvement

            service_reduced = service * f_improvement

            return service_reduced, service_reduced_techs
    except KeyError:
        return service, service_techs

def apply_scenario_drivers(
        enduse,
        sector,
        fuel_y,
        dw_stock,
        region,
        gva_industry,
        gva_per_head,
        population,
        reg_scen_drivers,
        base_yr,
        curr_yr
    ):
    """The fuel data for every end use are multiplied
    with respective scenario drivers. This is either
    done based on a dwelling stock or not.

    Arguments
    ----------
    enduse: str
        Enduse
    sector : str
        sector
    fuel_y : array
        Yearly fuel per fueltype
    dw_stock : object
        Dwelling stock
    region : str
        Region name
    gva_industry : dict
        GVA
    gva_per_head : dict
        GVA
    population : dict
        Population
    reg_scen_drivers : dict
        Scenario drivers per enduse
    base_yr : int
        Base year
    curr_yr : int
        Current year

    Returns
    -------
    fuel_y : array
        Changed yearly fuel per fueltype
    """
    if not dw_stock: # No dwelling stock is available
        #Calculate non-dwelling related scenario drivers
        scenario_drivers = reg_scen_drivers[enduse]

        by_driver, cy_driver = 1, 1

        for scenario_driver in scenario_drivers:

            # Get correct data depending on driver
            if scenario_driver == 'gva':
                if not sector:
                    by_driver_data = gva_per_head[base_yr][region]
                    cy_driver_data = gva_per_head[curr_yr][region]
                else:
                    try:
                        gva_sector_lu = lookup_tables.economic_sectors_regional_MISTRAL()
                        gva_sector_nr = gva_sector_lu[sector]['match_int']

                        # Get sector specific GVA
                        by_driver_data = gva_industry[base_yr][gva_sector_nr][region]
                        cy_driver_data = gva_industry[curr_yr][gva_sector_nr][region]

                    except KeyError:
                        # No sector specific GVA data defined
                        logging.debug("No gva data found for sector {} {} ".format(sector, curr_yr))
                        by_driver_data = gva_per_head[base_yr][region]
                        cy_driver_data = gva_per_head[curr_yr][region]

                    #logging.info("gva  {}   {}".format(by_driver_data, cy_driver_data))

            elif scenario_driver == 'population':
                by_driver_data = population[base_yr][region]
                cy_driver_data = population[curr_yr][region]

                #logging.info("pop  {}   {}".format(population[base_yr][region], population[curr_yr][region]))

            # Multiply drivers
            by_driver *= by_driver_data
            cy_driver *= cy_driver_data

            # Testing
            if math.isnan(by_driver_data) or math.isnan(cy_driver_data):
                raise Exception("Scenario driver error")

        # Calculate scenario driver factor
        try:
            factor_driver = cy_driver / by_driver
        except ZeroDivisionError:
            factor_driver = 1

        fuel_y = fuel_y * factor_driver
    else:
        #Scenario driver calculation based on dwelling stock

        # Test if enduse has a dwelling related scenario driver
        if hasattr(dw_stock[base_yr], enduse) and curr_yr != base_yr:

            # Scenariodriver of dwelling stock base year and new stock
            by_driver = getattr(dw_stock[base_yr], enduse)
            cy_driver = getattr(dw_stock[curr_yr], enduse)

            # Calculate scenario driver factor
            try:
                factor_driver = cy_driver / by_driver
            except ZeroDivisionError:
                factor_driver = 1

            # Testing
            if math.isnan(factor_driver):
                raise Exception("Scenario driver error")

            fuel_y = fuel_y * factor_driver
        else:
            pass #enduse not define with scenario drivers

    return fuel_y

def generic_demand_change(
        enduse,
        sector,
        fuel_y,
        generic_enduse_change,
        curr_yr
    ):
    """Calculates fuel based on assumed overall enduse or enduse/sector specific
    fuel consumption changes.

    The changes are assumed across all fueltypes.
    Because for enduses where no technologies are defined, a linear
    diffusion is suggested to best represent multiple sigmoid efficiency
    improvements of individual technologies.

    Arguments
    ----------
    enduse : str
        Enduse
    sector : str
        Sector. If True, then the change is valid for all sectors. if None, then no sector is defined for enduse.
    fuel_y : array
        Yearly fuel per fueltype
    generic_enduse_change : dict
        Change in overall enduse for every enduse (percent ey)
    curr_yr : int
        Current year

    Returns
    -------
    fuel_y : array
        Yearly new fuels
    """
    change_enduse = False

    # True for all sectors or no sector defined
    if sector is True or sector is None:
        try:
            change_cy = generic_enduse_change[enduse][curr_yr]
            change_enduse = True
        except KeyError: #no sector enduse is defined
            pass
    else:
        try:
            # Generic enduse change is defined sector specific
            change_cy = generic_enduse_change[enduse][sector][curr_yr]
            change_enduse = True
        except (KeyError, TypeError):
            pass
        try:
            # Generic enduse change is not defined sector specific
            change_cy = generic_enduse_change[enduse][curr_yr]

            change_enduse = True
        except (KeyError, TypeError):
            pass

    if change_enduse is True and change_cy != 0:
        fuel_y = fuel_y * (1 + change_cy)
    else:
        #logging.debug("No annual parameters are provided for enduse %s %s", enduse, sector)
        pass

    return fuel_y

def apply_weather_correction(
        enduse,
        fuel_y,
        cooling_factor_y,
        heating_factor_y,
        enduse_space_heating,
        enduse_space_cooling
    ):
    """Change fuel demand for heat and cooling service
    depending on changes in HDD and CDD within a region
    (e.g. climate change induced). Change fuel consumption based on
    climate change induced temperature differences

    Arguments
    ----------
    enduse : str
        Enduse
    fuel_y : array
        Yearly fuel per fueltype
    cooling_factor_y : array
        Distribution of fuel within year to days (yd)
    heating_factor_y : array
        Distribution of fuel within year to days (yd)
    enduse_space_heating : list
        Enduses defined as space heating
    enduse_space_cooling : list
        Enduses defined as space cooling

    Return
    ------
    fuel_y : array
        Changed yearly fuel per fueltype

    Note
    ----
    - `cooling_factor_y` and `heating_factor_y` are based on the sum
        over the year. Therefore it is assumed that fuel correlates
        directly with HDD or CDD.
    """
    if enduse in enduse_space_heating:
        fuel_y = fuel_y * heating_factor_y
    elif enduse in enduse_space_cooling:
        fuel_y = fuel_y * cooling_factor_y

    return fuel_y

def apply_smart_metering(
        enduse,
        fuel_y,
        sm_assump,
        strategy_vars,
        curr_yr
    ):
    """Calculate fuel savings depending on smart meter penetration.
    Change fuel consumption based on smart meter induced general savings.

    Arguments
    ----------
    enduse : str
        Enduse
    fuel_y : array
        Yearly fuel per fueltype
    sm_assump : dict
        smart meter assumptions
    strategy_vars : dict
        Base simulation parameters

    Returns
    -------
    fuel_y : array
        New fuel per year

    Note
    -----
    - The smart meter penetration is assumed with a sigmoid diffusion.

    - In the assumptions the maximum penetration and also the
        generally fuel savings for each enduse can be defined.
    """
    # Enduse saving potentail of enduse of smart meter
    enduse_savings = sm_assump['savings_smart_meter'][enduse]

    # Smart meter penetration in current year (percentage of people having smart meters)
    penetration_cy = strategy_vars['smart_meter_p'][curr_yr]

    # Smart meter penetration in base year (percentage of people having smart meters)
    penetration_by = sm_assump['smart_meter_p_by']

    # Calculate fuel savings
    saved_fuel = fuel_y * (penetration_cy - penetration_by) * enduse_savings

    # Substract fuel savings
    fuel_y = fuel_y - saved_fuel

    return fuel_y

def convert_service_to_p(tot_s_y, s_fueltype_tech):
    """Calculate fraction of service for every technology
    of total service

    Arguments
    ----------
    tot_s_y : float
        Total yearly service
    s_fueltype_tech : dict
        Service per technology and fueltype

    Returns
    -------
    s_tech_p : dict
        All tecnology services are
        provided as a fraction of total service

    Note
    ----
    Iterate over values in dict and apply calculations
    """
    if tot_s_y == 0:
        _total_service = 0
    else:
        _total_service = 1 / tot_s_y

    # Iterate all technologies and calculate fraction of total service
    s_tech_p = {}
    for tech_services in s_fueltype_tech.values():
        for tech, service_tech in tech_services.items():
            s_tech_p[tech] = _total_service * service_tech

    return s_tech_p

def get_service_diffusion(param_tech, curr_yr):
    """Calculate energy service fraction of technologies
    with increased service for current year based
    on sigmoid diffusion or linear diffusion according
    to provided param_tech

    Arguments
    ----------
    param_tech : dict
        Sigmoid diffusion parameters per technology
    curr_yr : dict
        Current year

    Returns
    -------
    s_tech_p : dict
        Share of service per technology of current year
    """
    if param_tech['l_parameter'] is None:
        s_tech_p = 0
    elif param_tech['l_parameter'] == 'linear':

        # Calcuate linear diffusion
        s_tech_p = param_tech['linear_slope'] * curr_yr + param_tech['linear_y_intercept']
    else:
        # Calculate sigmoid diffusion
        s_tech_p = diffusion_technologies.sigmoid_function(
            curr_yr,
            param_tech['l_parameter'],
            param_tech['midpoint'],
            param_tech['steepness'])

    # becuase of rounding error may be minus
    if s_tech_p < 0:
        assert s_tech_p < 0.0001 # Trow error if not rounding error. Something else went wrong
        s_tech_p = 0

    return s_tech_p

def apply_service_switch(
        enduse,
        s_tech_y_cy,
        all_technologies,
        curr_yr,
        base_yr,
        sector,
        annual_tech_diff_params,
        crit_switch_happening
    ):
    """Apply change in service depending on defined service switches.

    The service which is newly fulfilled by new technologies (as defined
    in the service switches) is substracted of the replaced
    technologies proportionally to their base year contribution.

    Arguments
    ---------
    enduse : str
        Enduse
    s_tech_y_cy : dict
        Service per technology
    all_technologies : dict
        Technologies to iterate
    curr_yr : int
        Current year
    base_yr : int
        Base year
    sector : str
        Sector
    annual_tech_diff_params : dict
        Sigmoid technology specific calculated annual diffusion values
    crit_switch_happening : bool
        Criteria wheter switch is defined or not

    Returns
    -------
    switched_s_tech_y_cy : dict
        Service per technology in current year after switch in a year
    """
    # ----------------------------------------
    # Test wheter switch is defined or not
    # ----------------------------------------
    crit_switch_service = fuel_service_switch.get_switch_criteria(
        enduse,
        sector,
        crit_switch_happening,
        base_yr,
        curr_yr)

    # ---------------------------------------
    # Calculate switch
    # ----------------------------------------
    if crit_switch_service:

        # Service of all technologies
        service_all_techs = sum(s_tech_y_cy.values())

        switched_s_tech_y_cy = {}
        for tech in all_technologies:
            # Get service share per tech of cy of sigmoid parameter calculations
            p_s_tech_cy = annual_tech_diff_params[tech][curr_yr]
            switched_s_tech_y_cy[tech] = service_all_techs * p_s_tech_cy
        return switched_s_tech_y_cy
    else:
        return s_tech_y_cy

def apply_cooling(
        enduse,
        fuel_y,
        cooled_floorarea,
        cooled_floorarea_p_by,
        curr_yr):
    """Apply changes for cooling enduses depending
    on assumption of how much of the floor area in percent
    is cooled

    It is aassumption a linear correlation between the
    percentage of cooled floor space (area) and energy demand.

    Arguments
    ---------
    enduse : str
        Enduse
    fuel_y : array
        Annual fuel demand
    cooled_floorarea : dict
        Strategy variables
    cooled_floorarea_p_by : dict
        Assumption about cooling floor area in base year
    curr_yr : int
        Current year

    Returns
    -------
    fuel_y : array
        Fuel array (either changed fuel depending on cooling percentage)
        of identical array
    """
    try:
        # Floor area share cooled in current year
        cooled_floorarea_p_cy = cooled_floorarea[enduse][curr_yr]

        # Calculate factor
        floorarea_cooling_factor = cooled_floorarea_p_cy / cooled_floorarea_p_by

        # Apply factor
        fuel_y = fuel_y * floorarea_cooling_factor
    except KeyError:
        pass

    return fuel_y

def generic_fuel_switch(
        enduse,
        sector,
        curr_yr,
        base_yr,
        fuel_switch,
        fuel_y
    ):
    """Generic fuel switch in an enduse (e.g. replacing a fraction
    of a fuel with another fueltype)

    Arguments
    ---------
    enduse : str
        Enduse
    sector : str
        Sector
    curr_yr : str
        Current year of simulation
    fuel_switch : str
        Fuel swithces
    fuel_y : str
        Fuel of specific enduse and sector

    Returns
    -------
    fuel_y : array
        Annual fuel demand per fueltype
    """
    if base_yr == curr_yr:
        return fuel_y

    try:
        # Test if switch is defined for sector
        fuel_switch = fuel_switch[enduse][sector]
        switch_defined = True
    except KeyError:
        switch_defined = False

        # Test is not a switch for the whole enduse (across every sector) is defined
        try:
            keys_of_switch = list(fuel_switch[enduse].keys())

            for key_of_switch in keys_of_switch:
                # Test wheter switches for sectors are provided
                if 'param_info' in fuel_switch[enduse][key_of_switch]: #one switch
                    fuel_switch = fuel_switch[enduse]
                    switch_defined = True
                else:
                    pass # Switch is not defined for this sector
        except KeyError:
            pass

    if switch_defined is True:
        for fueltype_new_int in fuel_switch.keys():

            if fuel_switch[fueltype_new_int][curr_yr] != 0:

                # Get fueltype to switch to (new)
                fueltype_new_int = int(fuel_switch[fueltype_new_int]['param_info']['fueltype_new'])

                fueltype_replace_int = int(fuel_switch[fueltype_new_int]['param_info']['fueltype_replace'])

                # Value of current year
                fuel_share_switched_cy = fuel_switch[fueltype_new_int][curr_yr]

                # Substract fuel
                fuel_minus = fuel_y[fueltype_replace_int] * (1 - fuel_share_switched_cy)
                fuel_y[fueltype_replace_int] -= fuel_minus

                # Add fuel
                fuel_y[fueltype_new_int] += fuel_minus
            else:
                pass # no fuel switch defined

    return fuel_y

def industry_enduse_changes(
        enduse,
        sector,
        curr_yr,
        strategy_vars,
        fuels,
        assumptions
    ):
    """This function changes the demand if the enduse
    is a an industrial enduse depending on assumed
    industry related scenario paramters

    Arguments
    ---------
    enduse : str
        Enduse
    sector : str
        Sector
    curr_yr : int
        Current year
    strategy_vars : dict
        All strategy variables
    fuels : array
        Annual fuels

    Returns
    --------
    fuels : np.array
        Changed fuels depending on scenario
    """
    if enduse == 'is_high_temp_process' and sector == 'basic_metals':

        # Calculate factor depending on fraction
        # of hot and cold steel rolling process
        factor = hot_cold_process(
            curr_yr,
            strategy_vars['p_cold_rolling_steel'],
            assumptions)

        fuels_out = fuels * factor

        return fuels_out
    else:
        return fuels

def hot_cold_process(
        curr_yr,
        p_cold_rolling_steel,
        assumptions
    ):
    """Calculate factor based on the fraction of hot
    and cold rolling processes in steel manufacturing.
    The fraction of either process is calculated based on
    the scenario input of the future share of cold rollling
    processes. A sigmoid diffusion towards this fucture defined
    fraction is implemented.

    Arguments
    ----------
    curr_yr : int
        Current year
    strategy_vars : dict
        Strategy variables
    assumptions : dict
        Assumptions including efficiencies of either process
        and the base year share

    Returns
    -------
    factor : float
        Factor to change energy demand
    """
    # Reduce demand depending on fraction of hot and cold steel rolling process
    p_cold_rolling_by = assumptions.p_cold_rolling_steel_by
    p_hot_rolling_by = 1.0 - p_cold_rolling_by

    # Current year fractions
    p_cold_rolling_cy = p_cold_rolling_steel[curr_yr]
    p_hot_rolling_cy = 1 - p_cold_rolling_cy

    # Efficiencies of processes
    eff_cold = assumptions.eff_cold_rolling_process
    eff_hot = assumptions.eff_hot_rolling_process

    p_by = p_cold_rolling_by * eff_cold + p_hot_rolling_by * eff_hot
    p_cy = p_cold_rolling_cy * eff_cold  + p_hot_rolling_cy * eff_hot

    factor = p_cy  / p_by

    return factor



def crit_dim_var(var):
    """Check if nested dict or not

    Arguments
    ---------
    var : dict
        Dictionary to test wheter single or multidimensional parameter

    Returns
    -------
    single_dimension : bool
        True: Single dimension, False: Multidimensional parameter
    """
    single_dimension = True

    # Test if list nor not
    if type(var) is list:
        for list_entry in var:
            for key, value in list_entry.items():

                if type(value) is not list:
                    if hasattr(value, 'keys') and key not in ['regional_vals_by', 'regional_vals_cy']:
                        if len(value.keys()) != 0:
                            single_dimension = False
                        else:
                            pass
    else:
       for key, value in var.items():
            if type(value) is not list:
                if hasattr(value, 'keys'):
                    if len(value.keys()) != 0 and key not in ['regional_vals_by', 'regional_vals_cy']:
                        single_dimension = False
            else:
                if value == []:
                    pass
                else:
                    single_dimension = False

    return single_dimension
    
#a = [{'base_yr': 2015, 'end_yr': 2050, 'fueltype_new': 0, 'sig_steepness': 1, 'value_by': 5, 'sig_midpoint': 0, 'regional_specific': False, 'value_ey': 5, 'diffusion_choice': 'linear', 'fueltype_replace': 0}]
#r = crit_dim_var(a)

a =  [
    {
        'sector': 'dummy_sector',
        'regional_specific': True,
        'value_ey': 1.0,
        'regional_vals_by': {'E06000001': 0.05, 'E06000018': 0.05}}]

r = crit_dim_var(a)
print(r)


"""Plotting styles
"""
from collections import OrderedDict
import palettable

def linestyles():
    """
    https://matplotlib.org/gallery/lines_bars_and_markers/linestyles.html

    """
    linestyles = OrderedDict(
        [
            ('solid',               (0, ())),
            ('loosely dotted',      (0, (1, 10))),
            ('dotted',              (0, (1, 5))),
            ('densely dotted',      (0, (1, 1))),

            ('loosely dashed',      (0, (5, 10))),
            ('dashed',              (0, (5, 5))),
            ('densely dashed',      (0, (5, 1))),

            ('loosely dashdotted',  (0, (3, 10, 1, 10))),
            ('dashdotted',          (0, (3, 5, 1, 5))),
            ('densely dashdotted',  (0, (3, 1, 1, 1))),

            ('loosely dashdotdotted', (0, (3, 10, 1, 10, 1, 10))),
            ('dashdotdotted',         (0, (3, 5, 1, 5, 1, 5))),
            ('densely dashdotdotted', (0, (3, 1, 1, 1, 1, 1)))])

    linestyles_out = {}
    for nr, (name, linestyle) in enumerate(linestyles.items()):
        linestyles_out[nr] = linestyle

    return linestyles_out

def marker_list():
    """markers
    """

    markers = [
        "o",	# circle
        "*",	# star
        "+",	# plus
        ".",	# point
        ",",	# pixel
        "P",	# plus (filled)
        "x",	# x
        "v",	# triangle_down
        "^",	# triangle_up
        "<",	# triangle_left
        ">",	# triangle_right
        "1",	# tri_down
        "2",	# tri_up
        "3",	# tri_left
        "4",	# tri_right
        "8",	# octagon
        "s",	# square
        "p",	# pentagon
        "h",	# hexagon1
        "H",	# hexagon2
        "X",	# x (filled)
        "D",	# diamond
        "d",	# thin_diamond
        "|",	# vline
    ]
    return markers

def color_list():
    """ List with colors
    """
    color_list = [
        'darkturquoise',
        'orange',
        'firebrick',
        'darkviolet',
        'khaki',
        'olive',
        'darkseagreen',
        'darkcyan',
        'indianred',
        'darkblue',
        'orchid',
        'gainsboro',
        'mediumseagreen',
        'lightgray',
        'mediumturquoise',
        'lemonchiffon',
        'cadetblue',
        'lightyellow',
        'lavenderblush',
        'coral',
        'purple',
        'aqua',
        'mediumslateblue',
        'darkorange',
        'mediumaquamarine',
        'darksalmon',
        'beige']

    return color_list

def get_colorbrewer_color(color_prop, color_palette, inverse=False):
    """
    Get hex color from colorbrewer

    Arguments
    ----------
    color_prop : str
        Sequential or qualitative color criteria
    color_palette : str
        Name of colorbrewer palette
    invers : default (False)
        Invert color criteria
    """
    if color_prop == 'sequential':
        color_list = getattr(palettable.colorbrewer.sequential, color_palette).hex_colors
    if color_prop == 'qualitative':
        color_list = getattr(palettable.colorbrewer.qualitative, color_palette).hex_colors

    if inverse:
        color_list = color_list[::-1]
    else:
        pass

    return color_list

def color_list_selection():
    """
    """
    color_list_selection = [
        'seagreen',
        'slateblue', #'silver',
        #'coral',

        'dimgray',
        'gray',
        'darkgray',
        'livghray',
        'silver',
        'gainsboro',
        'whitesmoke',

        'steelblue',
        'tomato',
        'slateblue',
        'moccasin', #darkturquoise',
        'khaki',
        'orange',
        'firebrick',
        'darkviolet',
        'olive',
        'darkseagreen',
        'darkcyan',
        'indianred',
        'darkblue'
        ]

    return color_list_selection

def rs_color_list_selection():
    """
    """
    color_list_selection = [
        'seagreen',
        'slateblue', #'silver',
        #'coral',

        'dimgray',
        'gray',
        'darkgray',
        'lightgray',
        'silver',
        'gainsboro',
        'whitesmoke',

        'steelblue',
        'tomato',
        #'slateblue',
        'moccasin', #darkturquoise',
        'khaki',
        'orange'
        ]

    return color_list_selection

def ss_color_list_selection():
    """
    """
    color_list_selection = [
        'seagreen',
        'slateblue', #'silver',
        #'coral',

        #'lightslategrey',
        'dimgray',
        'gray',
        #'azure',
        'darkgray',
        'lightgray',
        'silver',
        'gainsboro',
        'whitesmoke',
        'gainsboro',
        'lightgray',
        


        'firebrick',
        'darkviolet',
        'olive',
        'darkseagreen',
        'darkcyan',
        'indianred',
        'darkblue',
        'slategrey'
        ]

    return color_list_selection

def is_color_list_selection():
    """
    """
    color_list_selection = [
        'seagreen',
        'slateblue',

        'dimgray',
        'gray',
        'darkgray',
        'lightgray',
        'silver',
        'whitesmoke',
        'gainsboro',
        

        'saddlebrown',
        'indigo',
        'cadetblue',
        'yellowgreen',
        'peachpuff',
        'tan',
        'pink']

    return color_list_selection

def color_list_selection_dm():
    """
    """
    color_list_selection = [
        'darkturquoise',
        'orange',
        'firebrick',
        'darkviolet',
        'khaki',
        'olive',
        'darkseagreen',
        'darkcyan',
        'indianred',
        'darkblue']

    color_list_selection = [
        '#83c17e', #lighggreen
        '#556b2f', #green
        '#3b6db0', # blue
        '#f8bf87', #brown
    ]

    return color_list_selection

def color_list_scenarios():
    """
    """
    color_list_selection = get_colorbrewer_color(
        color_prop='qualitative', #sequential
        color_palette='Accent_6', #'Set3_12',
        inverse=False) # #https://jiffyclub.github.io/palettable/colorbrewer/sequential/'''

    color_list_selection = [
        '#7FC97F', #green
        'darkolivegreen', #'#BEAED4', #violet
        '#386CB0', #blue
        '#FDC086', #orange
        '#F0027F', #pink
        '#FFFF99'  #yello
        ]
    
    color_list_selection = [
        '#83c17e', #lighggreen
        '#556b2f', #green
        '#3b6db0', # blue
        '#f8bf87', #brown
    ]
    '''
    # Set3_12
    color_list_selection = [
        'darkolivegreen',
        'blue',
        '#B3DE69', #'#FCCDE5',
        '#FB8072',  #'#FDB462',
        '#BEBADA', #blue #'#80B1D3',
        '#BC80BD', #pink
        '#8DD3C7',
        '#FFED6F', #yellow
        '#FFFFB3', #yellow
        '#D9D9D9', #grey
        '#BEBADA', #blue
        '#CCEBC5',
        'forestgreen',
        'rosybrown',
        'blue',
        'firebrick']
    '''
    return color_list_selection

def font_info(
        family='arial',
        color='black',
        weight='normal',
        size=8
    ):
    """
    """
    font_additional_info = {
        'family': 'arial',
        'color': 'black',
        'weight': 'normal',
        'size': size}

    return font_additional_info
